from .base import ObjectWithTemplate


class Verbose(ObjectWithTemplate):
    template = 'verbose'

    def __init__(self,
                 physics=0, cuts=0,
                 sd=0,
                 actions=0, actor=0, step=0,
                 error=0, warning=0,
                 output=0,
                 beam=0,
                 volume=0, image=0, geometry=0, core=0,
                 run=0, event=0, tracking=0):
        self.physics = physics
        self.cuts = cuts
        self.sd = sd
        self.actions = actions
        self.actor = actor
        self.step = step
        self.error = error
        self.warning = warning
        self.output = output
        self.beam = beam
        self.volume = volume
        self.image = image
        self.geometry = geometry
        self.core = core
        self.run = run
        self.event = event
        self.tracking = tracking


class MaterialDatabase(ObjectWithTemplate):
    template = 'misc/database'

    def __init__(self, path: str=None):
        self.path = path


class MaterialDatabaseLocal(MaterialDatabase):
    def __init__(self):
        super().__init__('./GateMaterials.db')

# TODO: add MaterialDatabaseShared


class Visualisation(ObjectWithTemplate):
    template = 'misc/visualisation'

    def __init__(self, is_disable=True):
        self.is_disable = is_disable



import click
from .initialize import generate
from ..utils.syscall import shell_call


@generate.group()
def mac():
    """
    Generate mac file.
    """


@mac.command()
@click.option('--target', '-t', help="Filename of script to run to generate mac file.")
@click.option('--config', '-c', help="config filename to generate macs.", default='main.mac')
@click.option('--output', '-o', help="MAC filename, will passed to script or predefined method.")
def script(target, output, config):
    """
    Generate mac file by running a .py file.
    """
    shell_call('python {} --output {} --config {}'.format(target, output, config))


@mac.command()
@click.option('--predefined', '-p', help="Name of predefined system to generate mac file.")
@click.option('--config', '-c', help="config filename to generate macs.")
@click.option('--target', '-t', help="MAC filename, will passed to script or predefined method.")
def predefined(target, config):
    """
    Generate mac file by predefined system.
    """
    raise NotImplementedError



from ..components.physics import *

# used in cylindricalPET ,ecat and multiPatchPET systems.


def pet_physics(cut_pair_list):
    return Physics([PhotoElectric(), Compton(), ElectronIonisation(),
                    RayleighScattering( models = (LivermoreModel(),)),
                    Bremsstrahlung(), PositronAnnihilation(),
                    RadioactiveDecay(), MultipleScattering('e+'), MultipleScattering('e-')], cut_pair_list)


def optical_physics(cut_pair_list):
    return Physics([OpticalAbsorption(), OpticalRayleigh(), OpticalBoundary(),
                    OpticalMie(), OpticalWLS(), Scintillation(),
                    PhotoElectric(), Compton(), RayleighScattering(),
                    ElectronIonisation(), Bremsstrahlung(),
                    EMultipleScattering('e+'), EMultipleScattering('e-')], cut_pair_list)


def gamma_physics(cut_pair_list):
    return Physics([PhotoElectric(), Compton(), RayleighScattering(),
                    ElectronIonisation(), Bremsstrahlung(),
                    EMultipleScattering('e+'), EMultipleScattering('e-')], cut_pair_list)


def spect_physics(cut_pair_list):
    return Physics([], cut_pair_list)



import click
from ..conf import config, KEYS, SUBMIT_KEYS
from .main import pygate
from typing import Iterable, Tuple


class Task:
    def __init__(self, broadcast=None, single=None):
        self.broadcast = broadcast
        self.single = single


def submit_kernel(tasks: Iterable[Task], subdir_patterns: Iterable[str], dryrun):
    from pygate.routine import submit
    d = submit.Directory('.')
    ops = []
    for t in tasks:
        if t.broadcast is not None:
            ops.append(submit.OpSubmitBroadcast(t.broadcast,
                                                subdir_patterns))
        if t.single is not None:
            ops.append(submit.OpSubmitSingleFile(t.single))
    r = submit.RoutineOnDirectory(d, ops, dryrun)
    r.work()
    return r.echo()


@pygate.command()
@click.option('--broadcast', '-b', multiple=True)
@click.option('--single', '-s', multiple=True)
def submit(broadcast, single):
    if len(broadcast) == 0 and len(single) == 0:
        submit_conf = config.get(KEYS.SUBMIT, {})
        broadcast = submit_conf.get(SUBMIT_KEYS.BROADCAST)
        single = submit_conf.get(SUBMIT_KEYS.SINGLE)
        tasks = [Task(b, s) for b, s in zip(broadcast, single)]
        click.echo(submit_kernel(tasks, config.get(KEYS.SUB_PATTERNS),
                                 config.get(KEYS.DRYRUN)))



from typing import Iterable


def shell_call(commands: Iterable[str] or str, is_echo=True):
    """
    Parameters:

    - `command`: Command to be executed.
    - `is_echo`: If True, print stdout and stderr of command to current stdout and stderr.
    """
    import subprocess
    import sys

    if isinstance(commands, str):
        commands = [commands]
    with subprocess.Popen(commands,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          shell=True) as p:
        out = p.stdout.read().decode()
        err = p.stderr.read().decode()
        if is_echo:
            sys.stdout.write(out)
            sys.stderr.write(err)
    return {'out': out, 'err': err}



from .base import OperationOnFile, RoutineOnDirectory
from dxl.fs import Directory, File
from enum import Enum

from pygate.analysis import predifined as pd


class KEYS:
    SOURCE = 'source'
    TARGET = 'target'
    ANALYSIS = 'analysis'


class AnalysisType(Enum):
    GAMMA_ENERGY_DEPOSIT_DISTRIBUTION = 'gamma_edep'


class OperationAnalysis(OperationOnFile):
    def __init__(self, source_csv_filename, target_h5_filename,
                 analysis_type: AnalysisType=None):
        super().__init__(target_h5_filename)
        self.source_filename = source_csv_filename
        if isinstance(analysis_type, str):
            self.analysis_type = AnalysisType(analysis_type)
        else:
            self.analysis_type = analysis_type

    def source(self, r: RoutineOnDirectory):
        return r.directory.attach_file(self.source_filename)

    def apply(self, r: RoutineOnDirectory):
        source = self.source(r)
        target = self.target(r)
        func_map = {
            AnalysisType.GAMMA_ENERGY_DEPOSIT_DISTRIBUTION: pd.gamma_energy_deposit_distribution
        }
        func_map[self.analysis_type](source.system_path(),
                                     target.system_path())
        return {KEYS.SOURCE: source.path.s,
                KEYS.TARGET: target.path.s,
                KEYS.ANALYSIS: self.analysis_type.value}

    def dryrun(self, r: RoutineOnDirectory):
        source = self.source(r)
        target = self.target(r)
        return {KEYS.SOURCE: source.path.s,
                KEYS.TARGET: target.path.s,
                KEYS.ANALYSIS: self.analysis_type.value}



from ..components.parameter import *
from ..components import parameter


def pet_parameters(acqu=AcquisitionPeriod(),
                   outputs=None,
                   rand=RandomEngine(seed='auto')):
    if outputs is None:
        outputs = [Root('pet', 1, 1, 1, None, 1), ]
    return Parameter(rand, outputs, acqu)


def optical_parameters(acqu=AcquisitionPrimaries(number=10000),
                       outputs=None,
                       rand=RandomEngineMersenneTwister(seed='auto')):
    if outputs is None:
        outputs = [Root('optical', 1, 1, None, 1, None), ]
    return Parameter(rand, outputs, acqu)


def optical_gamma(nb_primaries):
    outputs = [parameter.Root('optical', 1, 1, None, 1, None), ]
    return parameter.Parameter(parameter.RandomEngineMersenneTwister(seed='auto'),
                               outputs,
                               parameter.AcquisitionPrimaries(nb_primaries))



"""
"""
from .base import Operation, OperationOnFile, OperationOnSubdirectories, OpeartionWithShellCall, RoutineOnDirectory
from dxl.fs import Directory, File
from typing import TypeVar, Iterable
from pygate.components.simulation import Simulation
from pygate.scripts.shell import Script
import rx


class KEYS:
    SUBDIRECTORIES = 'subdirectories'
    TARGET = 'target'
    IS_TO_BROADCAST = 'is_to_broadcast'
    CONTENT = 'content'
    TO_BROADCAST_FILES = 'to_broadcast_files'


class TargetFileWithContent:
    def __init__(self, target: TypeVar('FileLike', File, str), is_to_broadcast: bool=True, content: str=None):
        if isinstance(target, File):
            self.target = target
        else:
            self.target = File(target)
        self.is_to_broadcast = is_to_broadcast
        self.content = content

    def to_dict(self):
        return {KEYS.TARGET: self.target.path.s,
                KEYS.IS_TO_BROADCAST: self.is_to_broadcast,
                KEYS.CONTENT: self.content}

    def save(self):
        if self.content is None:
            raise TypeError("None content to save.")
        self.target.save(self.content)


class OpAddToBroadcastFile(OperationOnFile):
    def __init__(self, filename):
        super().__init__(filename)

    def apply(self, r: RoutineOnDirectory):
        return self.dryrun(r)

    def dryrun(self, r: RoutineOnDirectory):
        return {KEYS.TARGET: self.target(r).path.s,
                KEYS.IS_TO_BROADCAST: True}


class OpGenerateFile(OperationOnFile):
    def __init__(self, filename: str, is_to_broadcast=True):
        super().__init__(filename)
        self.is_to_broadcast = is_to_broadcast

    def content(self, r: RoutineOnDirectory) -> str:
        raise NotImplementedError

    def target_file_with_content(self, r: RoutineOnDirectory):
        return TargetFileWithContent(self.target(r), self.is_to_broadcast,
                                     self.content(r))

    def apply(self, r: RoutineOnDirectory):
        f = self.target_file_with_content(r)
        f.save()
        return f.to_dict()

    def dryrun(self, r: RoutineOnDirectory):
        return self.target_file_with_content(r).to_dict()


class OpGeneratorMac(OpGenerateFile, OpeartionWithShellCall):
    def __init__(self, script_filename: str, mac_config: str=None, mac_filename: str=None):
        super.__init__(filename)
        self.mac_config = mac_config
        self.mac_filename = mac_filename

    def call_args(self, r: RoutineOnDirectory):
        result = ['python', self.target(r).path.s]
        if self.mac_config is not None:
            result += ['--config', self.mac_config]
        if self.mac_filename is not None:
            result += ['--target', self.mac_filename]
        return result


class OpGenerateMacTemplate(OpGenerateFile):
    def __init__(self, filename):
        super().__init__(filename)
        from ..scripts.shell import ScriptMacTemplate
        self.script = ScriptMacTemplate()

    def content(self, r: RoutineOnDirectory) -> str:
        return self.script.render()


class OpGeneratorShell(OperationOnFile):
    def __init__(self, filename: str, script: Script):
        super.__init__(filename)
        self.script = script

    def content(self, r: RoutineOnDirectory) -> str:
        return self.script.render()


class OpGeneratorPhantom(OperationOnFile):
    def __init__(self, filename: str):
        super.__init__(filename)


class OpSubdirectoriesMaker(Operation):
    def __init__(self, nb_split: int, subdirectory_format: str="sub.{}"):
        self.nb_split = nb_split
        self.fmt = subdirectory_format

    def apply(self, r: RoutineOnDirectory):
        result = self.dryrun(r)
        for n in result[KEYS.SUBDIRECTORIES]:
            r.directory.makedir(n)
        return result

    def dryrun(self, r: RoutineOnDirectory):
        return {KEYS.SUBDIRECTORIES: tuple([self.fmt.format(i) for i in range(self.nb_split)])}


class OpBroadcastFile(OperationOnSubdirectories):
    def files_to_broadcast(self, r: RoutineOnDirectory):
        result = []
        for res in r.result:
            if res.get(KEYS.IS_TO_BROADCAST, False):
                result.append(File(res[KEYS.TARGET], r.directory.filesystem))
        return result

    def dryrun(self, r: RoutineOnDirectory):
        dirs = self.subdirectories(r).to_list().to_blocking().first()
        dirs = [d.path.s for d in dirs]
        files = self.files_to_broadcast(r)
        files = [f.path.s for f in files]
        return {KEYS.SUBDIRECTORIES: dirs,
                KEYS.TO_BROADCAST_FILES: files}

    def apply(self, r: RoutineOnDirectory):
        result = self.dryrun(r)
        files = self.files_to_broadcast(r)
        (self.subdirectories(r).map(lambda d: d.sync(files))
         .to_list().to_blocking().first())
        return result



from typing import List
from .base import ObjectWithTemplate
from .utils import Vec3


class Source(ObjectWithTemplate):
    template = 'source/source'

    def __init__(self, name,
                 particle=None, activity=None, angle=None,
                 shape=None, position: Vec3=None):
        """
        Args:
            activity: str | int | float | list | tuple:
                str: directly attached
                int, float: add default unit becquerel
                list, tuple: [0]: value, [1]: unit
        """
        self.name = name
        self.particle = self.bind(particle)
        self.angle = self.bind(angle)
        self.shape = self.bind(shape)
        self.activity = self.unified_activity(activity)
        self.position = position

    def unified_activity(self, activity):
        if activity is None:
            return None
        if isinstance(activity, str):
            return activity
        if isinstance(activity, (int, float)):
            unit = 'becquerel'
        if isinstance(activity, (list, tuple)):
            unit = activity[1]
            activity = activity[0]
            if unit.lower() in ['bq']:
                unit = 'becquerel'
        return "{} {}".format(activity, unit)

    def bind(self, obj):
        if obj is not None:
            obj.src = self
        return obj

    def is_voxelized(self):
        return isinstance(self.shape, Voxelized)


class Particle(ObjectWithTemplate):
    template = 'source/particle/particle'
    particle_type = None

    def bind_source(self, source):
        self.src = source
        return self

    def __init__(self, unstable=None, halflife=None):
        self.scr = None
        self.unstable = unstable
        self.halflife = halflife


class ParticlePositron(Particle):
    template = 'source/particle/positron'
    particle_type = 'e+'

    def __init__(self, unstable=True, halflife=6586):
        super().__init__(unstable, halflife)


class ParticleGamma(Particle):
    template = 'source/particle/gamma'
    particle_type = 'gamma'

    def __init__(self,
                 unstable=True, halflife=6586.2,
                 monoenergy=511, back2back=True):
        super().__init__(unstable, halflife)
        self.back2back = back2back
        self.monoenergy = monoenergy


class Angular(ObjectWithTemplate):
    template = 'source/angular/angular'
    ang_type = None


class AngularISO(Angular):
    template = 'source/angular/iso'
    ang_type = 'iso'

    def __init__(self, ang=[0, 180, 0, 360]):
        self.ang = ang


class Shape(ObjectWithTemplate):
    template = 'source/shape/shape'
    shape = None

    def __init__(self, dimension):
        self.dimension = dimension


class ShapePlane(Shape):
    def __init__(self):
        super().__init__(dimension='Plane')


class ShapeSurfaceOrVolume(Shape):
    def __init__(self, dimension):
        if not dimension in ('Surface', 'Volume'):
            raise ValueError(
                'Invalid dimension {} for {}.'.format(dimension, __class__))
        super().__init__(dimension=dimension)


class Voxelized(Shape):
    template = 'source/shape/voxelized'
    shape = 'Voxelized'

    def __init__(self, read_table, read_file,
                 reader='interfile', translator='range',
                 position=None):
        self.read_table = read_table
        self.read_file = read_file
        self.reader = reader
        self.translator = translator
        self.position = position


class Cylinder(ShapeSurfaceOrVolume):
    template = 'source/shape/cylinder'
    shape = 'Cylinder'

    def __init__(self, radius, halfz, dimension):
        super().__init__(dimension)
        self.radius = radius
        self.halfz = halfz


class Sphere(Shape):
    template = 'source/shape/sphere'
    shape = 'Sphere'

    def __init__(self, radius, dimension):
        super().__init__(dimension)
        self.radius = radius


class Ellipsoid(Shape):
    template = 'source/shape/ellipsoid'
    shape = 'Ellipsoid'

    def __init__(self, half_size: Vec3, dimension):
        super().__init__(dimension)
        self.half_size = half_size


class Circle(ShapePlane):
    template = 'source/shape/circle'
    shape = 'Circle'

    def __init__(self, radius):
        super().__init__()
        self.radius = radius


class Annulus(ShapePlane):
    template = 'source/shape/annulus'
    shape = 'Annulus'

    def __init__(self, radius0, radius):
        super().__init__()
        self.radius = radius
        self.radius0 = radius0


class Ellipse(ShapePlane):
    template = 'source/shape/ellipse'
    shape = 'Ellipse'

    def __init__(self, half_size):
        super().__init__()
        self.half_size = half_size


class Rectangle(ShapePlane):
    template = 'source/shape/rectangle'
    shape = 'Rectangle'

    def __init__(self, half_size):
        super().__init__()
        self.half_size = half_size


class SourceList(ObjectWithTemplate):
    template = 'source/source_list'

    def __init__(self, sources: List[Source]):
        self.sources = sources



from ...base import ObjectWithTemplate


class System(ObjectWithTemplate):
    template = 'geometry/system'
    name = None
    attach_systems = None

    def __init__(self):
        self.levels = self.attach_systems()
        self.levels = {k: self.levels[k]
                       for k in self.levels if self.levels[k] is not None}


class PETscanner(System):
    name = 'PETscanner'

    def __init__(self, level1, level2, level3, level4, level5, sensitive_detectors=None):
        self.attach_systems = lambda: {
            'level1': level1,
            'level2': level2,
            'level3': level3,
            'level4': level4,
            'level5': level5
        }
        super().__init__(sensitive_detectors)


class Ecat(System):
    name = 'ecat'

    def __init__(self, block=None, crystal=None):
        self.attach_systems = lambda:  {'block': block, 'crystal': crystal}
        super().__init__()
        



class CylindricalPET(System):
    name = 'cylindricalPET'

    def __init__(self,
                 rsector=None, module=None, submodule=None, crystal=None,
                 layer0=None, layer1=None, layer2=None, layer3=None):
        self.attach_systems = lambda: {
            'rsector': rsector,
            'module': module,
            'submodule': submodule,
            'crystal': crystal,
            'layer0': layer0,
            'layer1': layer1,
            'layer2': layer2,
            'layer3': layer3,
        }
        super().__init__()

class MultiPatchPET(System):
    name = 'multiPatchPET'
    def __init__(self, container, patch_list):
        self.attach_systems = lambda: {'contianer':container, 
        'patch_list':patch_list}
        super().__init__()

class SPECThead(System):
    name = 'SPECThead'
    def __init__(self, crystal, pixel = None):
        self.attach_systems = lambda: {'crystal': crystal, 'pixel':pixel }
        super().__init__()

class OpticalSystem(System):
    name = 'OpticalSystem'
    def __init__(self,crystal, pixel = None):
        self.attach_systems = lambda: {'crystal': crystal, 'pixel':pixel}
        super().__init__()



from pathlib import Path

def get_template_path():
    from pathlib import Path
    from . import templates
    return str(Path(templates.__file__).parent.absolute())

configs = {
    'TEMPLATES_PATH': get_template_path()
}


from .base import ObjectWithTemplate
from .geometry import Volume
from typing import TypeVar as TV
from typing import List, Dict


class Model(ObjectWithTemplate):
    name = None
    template = 'physics/model'

    def __init__(self, particle=None):
        self.particle = particle
        self.process = None


class PenelopeModel(Model):
    name = 'PenelopeModel'


class StandardModel(Model):
    name = 'StandardModel'

class LivermoreModel(Model):
    name = 'LivermoreModel'

class LivermorePolarizedModel(Model):
    name = 'LivermorePolarizedModel'




class PhysicsProcess(ObjectWithTemplate):
    name = None
    template = 'physics/process'

    def __init__(self, models=None):
        if models is None:
            models = (StandardModel(),)
        self.models = models
        for m in self.models:
            m.process = self

    def content_in_adding(self):
        return self.name


class PhotoElectric(PhysicsProcess):
    name = 'PhotoElectric'


class Compton(PhysicsProcess):
    name = 'Compton'

class GammaConversion(PhysicsProcess):
    name = 'GammaConversion'

class RayleighScattering(PhysicsProcess):
    name = 'RayleighScattering'

    def __init__(self, models=(PenelopeModel(),)):
        super().__init__(models)


class ElectronIonisation(PhysicsProcess):
    name = 'ElectronIonisation'

    def __init__(self, models=(StandardModel('e-'),
                               StandardModel('e+'))):
        super().__init__(models)


class Bremsstrahlung(PhysicsProcess):
    name = 'Bremsstrahlung'

    def __init__(self, models=(StandardModel('e-'),
                               StandardModel('e+'))):
        super().__init__(models)


class PhysicsProcessWithoutModels(PhysicsProcess):
    def __init__(self):
        super().__init__(tuple())



class PositronAnnihilation(PhysicsProcessWithoutModels):
    name = 'PositronAnnihilation'

class RadioactiveDecay(PhysicsProcessWithoutModels):
    name = 'RadioactiveDecay'

class OpticalAbsorption(PhysicsProcessWithoutModels):
    name = 'OpticalAbsorption'


class OpticalRayleigh(PhysicsProcessWithoutModels):
    name = 'OpticalRayleigh'


class OpticalBoundary(PhysicsProcessWithoutModels):
    name = 'OpticalBoundary'


class OpticalMie(PhysicsProcessWithoutModels):
    name = 'OpticalMie'


class OpticalWLS(PhysicsProcessWithoutModels):
    name = 'OpticalWLS'


class Scintillation(PhysicsProcessWithoutModels):
    name = 'Scintillation'


class MultipleScattering(PhysicsProcessWithoutModels):
    name = 'MultipleScattering'

    def __init__(self, particle):
        super().__init__()
        self.p = particle

    def content_in_adding(self):
        return "{} {}".format(self.name, self.p)

class EMultipleScattering(PhysicsProcessWithoutModels):
    name = 'eMultipleScattering'
    
    def __init__(self, particle):
        super().__init__()
        self.p = particle

    def content_in_adding(self):
        return "{} {}".format(self.name, self.p)

class PhysicsList(ObjectWithTemplate):
    template = 'physics/list'

    def __init__(self, physics_processes: List[PhysicsProcess]):
        self.pps = physics_processes


class Cuts(ObjectWithTemplate):
    template = 'physics/cuts'
    STD_PARTICLES = ('Gamma', 'Electron', 'Positron')

    def __init__(self, volume: Volume, cuts: TV('CutT', float, Dict[Volume, float]), max_step: float=None):
        """
        All units are mm.
        cuts: map from partical name to cut.
        """
        self.volume = volume
        if isinstance(cuts, float):
            cuts = {k: cuts for k in self.STD_PARTICLES}
        self.cuts = cuts
        self.max_step = max_step


class Physics(ObjectWithTemplate):
    template = 'physics/physics'

    def __init__(self, physics_list, cuts_list):
        self.pl = physics_list
        self.cl = cuts_list



from typing import Iterable, Dict, Any

from dxl.fs import Directory, File


class Operation:
    def apply(self, routine):
        raise NotImplementedError

    def dryrun(self, routine):
        raise NotImplementedError


class Routine:
    def __init__(self, operations: Iterable[Operation]=(),
                 dryrun=False, verbose=0):
        self.dryrun = dryrun
        self.ops = operations
        self.verbose = verbose
        self.result = []

    def last_result(self):
        if len(self.result) == 0:
            return {}
        return self.result[-1]

    def work(self):
        for o in self.ops:
            if self.dryrun:
                self.result.append(o.dryrun(self))
            else:
                self.result.append(o.apply(self))
        return tuple(self.result)

    def echo(self):
        import json
        return json.dumps(self.result, sort_keys=True, indent=4,
                          separators=(',', ': '))


class RoutineOnDirectory(Routine):
    def __init__(self,
                 directory: Directory,
                 operations: Iterable[Operation]=(),
                 dryrun=False, verbose=0):
        super().__init__(operations, dryrun, verbose)
        self.directory = directory

    # def list_matched_dirs(self, pattern):
    #     return (self.directory.listdir_as_observable()
    #             .filter(match_directory(pattern)))

    # def list_matched_files(self, pattern):
    #     from dxl.fs import match_file
    #     return (self.directory.listdir_as_observable()
    #             .filter(match_file(pattern)))

    # def list_files_in_dirs(self, sub_dir_pattern, filename):
    #     return self.list_matched_dirs().map(lambda d: d.attach(filename))


class OperationOnFile(Operation):
    """
    Methods:

    - `self.target(r: RoutineOnDirectory)`: Return corresponding File object.
    - `self.dryrun(r: RoutineOnDirectory)`: Return Dict['target', system path].
    """

    def __init__(self, filename: str):
        self.filename = filename

    def target(self, r: RoutineOnDirectory)-> File:
        return r.directory.attach_file(self.filename)

    def dryrun(self, r: RoutineOnDirectory) -> Dict[str, Any]:
        return {'target': self.target(r).path.s}

    def apply(self, r: RoutineOnDirectory) -> Dict[str, Any]:
        return self.dryrun(r)


class OperationOnSubdirectories(Operation):
    def __init__(self, patterns: Iterable[str]):
        self.patterns = patterns

    def subdirectories(self, r: RoutineOnDirectory) -> 'Observable[Directory]':
        from dxl.fs import match_directory
        return (r.directory.listdir_as_observable()
                .filter(match_directory(self.patterns)))


class OpeartionWithShellCall(Operation):
    def call_args(self, r: Routine):
        raise NotImplementedError

    def stdout(self, r: Routine):
        return None

    def run_child_program(self, r: Routine):
        import subprocess
        import sys

        with subprocess.Popen(self.call_args(r),
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              shell=True) as p:
            out = p.stdout.read().decode()
            err = p.stderr.read().decode()
            sys.stdout.write(out)
            sys.stderr.write(err)
        return {'out': out, 'err': err}

    def apply(self, r: Routine):
        result = self.dryrun(r)
        result.update(self.run_child_program(r))
        return result

    def dryrun(self, r: Routine) -> Dict[str, Iterable[str]]:
        return {'call_args': self.call_args(r)}



class EnvironmentOfPackage:
    def __init__(self, pkg_name):
        self._env = None
        self.name = pkg_name

    def get_or_create_env(self):
        from jinja2 import Environment, PackageLoader
        if self._env is None:
            self._env = Environment(loader=PackageLoader(self.name))
            self._env.filters['default_suffix'] = default_suffix
            self._env.filters['ds'] = default_suffix
        return self._env


"""
Custom filters
"""


def default_suffix(s: str, suffix: str, is_auto_add_point_prefix=True):
    if is_auto_add_point_prefix and not suffix.startswith('.'):
        suffix = '.' + suffix
    if not s.endswith(suffix):
        s += suffix
    return s


class ObjectWithTemplateBase:
    pkg_env = None
    template = None

    def template_name(self):
        template_name = self.template
        if not template_name.endswith('.j2'):
            template_name += '.j2'
        return template_name

    def render(self):
        template = (self.pkg_env.get_or_create_env()
                    .get_template(self.template_name()))
        return template.render(o=self)



from .base import Operation, OperationOnFile, OperationOnSubdirectories
from dxl.fs import Directory, File
from typing import Callable, Iterable, Dict, Any
from .base import RoutineOnDirectory
from dxl.cluster import submit_slurm

# TODO: Rework dicts to object, use ToSubmit Object replacing those dicts


class KEYS:
    SUBMITTED = 'submitted'
    WORK_DIR = 'work_directory'
    SCRIPT_FILE = 'script_file'
    SID = 'sid'
    DEPENDENCIES = 'depens'


def depens_from_result_dict(r: Dict[str, Any]) -> Iterable[int]:
    try:
        result = []
        if KEYS.SUBMITTED in r:
            for t in r[KEYS.SUBMITTED]:
                if t.get(KEYS.SID) is not None:
                    result.append(int(t[KEYS.SID]))
        return result
    except Exception as e:
        raise ValueError("Failed to parse depens from result {}.".format(r))


def append_depens_to_dict(to_submit: Dict[str, Any], previous_result: Dict[str, Any]):
    result = dict(to_submit)
    result[KEYS.DEPENDENCIES] = depens_from_result(previous_result)
    return result


def parse_paths_from_dict(r: Dict[str, Any]) -> Dict[str, Any]:
    try:
        result = dict(r)
        for k in [KEYS.WORK_DIR, KEYS.SCRIPT_FILE]:
            if k in r:
                result[k] = r[k].path.s
        for k in [KEYS.SID, KEYS.DEPENDENCIES]:
            if k in r:
                result[k] = r[k]
        return result
    except Exception as e:
        raise ValueError("Failed to parse paths from result {}.".format(r))


def submit_from_dict(r: Dict[str, Any]) -> Dict[str, Any]:
    sid = submit_slurm(r[KEYS.WORK_DIR],
                       r[KEYS.SCRIPT_FILE],
                       r.get(KEYS.DEPENDENCIES, ()))

    result = dict(r)
    result[KEYS.SID] = sid
    return result


class OpSubmitBroadcast(OperationOnSubdirectories, OperationOnFile):
    """
    Submit all files with given filename in subdirectories.
    """

    def __init__(self, filename, subdirectory_patterns: Iterable[str]):
        OperationOnSubdirectories.__init__(self, subdirectory_patterns)
        OperationOnFile.__init__(self, filename)

    def to_submit(self, r: RoutineOnDirectory) -> 'Observable[Dict[str, Any]]':
        return (self.subdirectories(r)
                .map(lambda d: {KEYS.WORK_DIR: d,
                                KEYS.SCRIPT_FILE: self.target(r)}))

    def apply(self, r: RoutineOnDirectory) -> Dict[str, Iterable[Dict[str, str]]]:
        result = (self.to_submit(r)
                  .map(submit_from_dict)
                  .map(parse_paths_from_dict)
                  .to_list().to_blocking().first())
        return {KEYS.SUBMITTED: result}

    def dryrun(self, r: RoutineOnDirectory) -> Dict[str, Iterable[Dict[str, str]]]:
        result = (self.to_submit(r)
                  .map(parse_paths_from_dict)
                  .to_list().to_blocking().first())
        return {KEYS.SUBMITTED: result}


class OpSubmitSingleFile(OperationOnFile):
    def __init__(self, filename: str):
        super().__init__(filename)

    def to_submit(self, r: RoutineOnDirectory):
        submit_dict = {KEYS.WORK_DIR: r.directory,
                       KEYS.SCRIPT_FILE: self.target(r)}
        submit_dict[KEYS.DEPENDENCIES] = depens_from_result_dict(
            r.last_result())
        return submit_dict

    def apply(self, r: RoutineOnDirectory) -> Dict[str, Any]:
        result = submit_from_dict(self.to_submit(r))
        return {KEYS.SUBMITTED: parse_paths_from_dict(result)}

    def dryrun(self, r: RoutineOnDirectory) -> Dict[str, Any]:
        return {KEYS.SUBMITTED: parse_paths_from_dict(self.to_submit(r))}



from typing import List


def ignore_empty_newlines(s):
    return '\n'.join([x for x in s.split('\n') if x != ''])


def join_multiple(s, to_join):
    return to_join.join([x for x in s.split(to_join) if x != ''])


def drop_comments(s, comment_prefix='#'):
    lines = s.split('\n')
    lines = [s for s in lines if not s.startswith(comment_prefix)]
    return '\n'.join(lines)


def rstrip(s):
    lines = s.split('\n')
    lines = [s.rstrip() for s in lines]
    return '\n'.join(lines)

def expand_tabs(s, number_of_spaces=4):
    return s.replace('\t', ' '*number_of_spaces)

def ignore_multiple_whitespaces(s: str):
    s = join_multiple(s, ' ')
    s = join_multiple(s, '\n')
    s = rstrip(s)
    s = drop_comments(s)
    return s


def assert_equal_ignoring_multiple_whitespaces(obj, first, second):
    return obj.assertEqual(ignore_multiple_whitespaces(first),
                           ignore_multiple_whitespaces(second))



import click
from .main import pygate
from typing import Iterable

from ..conf import config, KEYS, SUBMIT_KEYS


class Task:
    def __init__(self, filename, method):
        self.filename = filename
        self.method = method


def merge_kernel(tasks: Iterable[Task], subdir_pattern: Iterable[str], dryrun):
    from pygate.routine import merger as merge
    d = merge.Directory('.')
    op_map = {
        'hadd': merge.OpMergeHADD,
        'cat': merge.OpMergeCat,
        'sum_bin': merge.OpMergeSumBinary,
        'catpdh5': merge.OpMergePandasConcatenate
    }
    ops = [op_map[t.method.lower()](t.filename, subdir_pattern) for t in tasks]
    r = merge.RoutineOnDirectory(d, ops, dryrun)
    r.work()
    return r.echo()


@pygate.command()
@click.option('--target', '-t', multiple=True)
@click.option('--method', '-m', multiple=True)
def merge(target, method):
    tasks = [Task(t, m) for t, m in zip(target, method)]
    click.echo(merge_kernel(tasks, config.get(KEYS.SUB_PATTERNS),
                            config.get(KEYS.DRYRUN)))



# import unittest
# # from fs.memoryfs import MemoryFS as FS
# from fs.tempfs import TempFS as FS
# from pygate import methods


# class TestMethods(unittest.TestCase):
#     def setUp(self):
#         self.fs = FS()
#         for i in range(5):
#             self.fs.makedir('sub.{}'.format(i))
#         for i in range(3):
#             self.fs.touch('test{}.mac'.format(i))
#         for i in range(3):
#             self.fs.touch('result{}.txt'.format(i))
#         self.fs.makedir('testdir')
#         self.fs.touch('testdir/tmp.txt')

#     def tearDown(self):
#         self.fs.close()

#     def test_files(self):
#         expected = set()
#         for i in range(3):
#             expected.add('test{}.mac'.format(i))
#             expected.add('result{}.txt'.format(i))
#         result = set()
#         methods.files(self.fs).subscribe(result.add)
#         self.assertEqual(result, expected)

#     def test_files_exclude(self):
#         expected = {'test{}.mac'.format(i) for i in range(3)}
#         result = set()
#         methods.files(self.fs, exclude_files=['result*']).subscribe(result.add)
#         self.assertEqual(result, expected)

#     def test_subdirs(self):
#         expected = {'sub.{}'.format(i) for i in range(5)}
#         result = set()
#         methods.dirs(self.fs, filters=['sub*']).subscribe(result.add)
#         self.assertEqual(result, expected)

#     def test_mapper_subdirs(self):
#         m = methods.Mapper(self.fs)
#         result = set()
#         m.files('result.txt').subscribe(result.add)
#         expected = {'sub.{}/result.txt'.format(i) for i in range(5)}
#         self.assertEqual(result, expected)



import click

from ..conf import config, KEYS, CLEAN_KEYS
from .main import pygate
from typing import Iterable


def clean_kernel(is_subdirectories, subdirectory_patterns: Iterable[str],
                 root_file_patterns: Iterable[str],
                 is_slurm_outputs, dryrun: bool):
    from pygate.routine import cleaner
    d = cleaner.Directory('.')
    ops = []
    if is_subdirectories:
        ops.append(cleaner.OpCleanSubdirectories(subdirectory_patterns))
    if len(root_file_patterns) > 0:
        ops.append(cleaner.OpCleanSource(root_file_patterns))
    if is_slurm_outputs:
        ops.append(cleaner.OpCleanFilesInAllDirectories(['*.out', '*.err'],
                                                        subdirectory_patterns))
    r = cleaner.RoutineOnDirectory(d, ops, dryrun)
    r.work()
    return r.echo()


@pygate.command()
@click.option('--subdirectories', '-d', help='remove subdirectories', is_flag=True)
@click.option('--root-files', '-f', help='remove files in work directory', multiple=True)
@click.option('--slurm-outputs', '-s', help='remove *.out *.err files', is_flag=True)
def clean(subdirectories, root_files, slurm_outputs):
    clean_conf = config.get(KEYS.CLEAN, {})
    if subdirectories is None:
        subdirectories = clean_conf[CLEAN_KEYS.IS_SUBDIRECTORIES]
    if slurm_outputs is None:
        slurm_outputs = clean_conf[CLEAN_KEYS.IS_SLURM_OUTPUTS]
    config_root_files = clean_conf[CLEAN_KEYS.ROOT_FILES]
    if len(config_root_files) > 0:
        root_files = tuple(list(root_files) + list(config_root_files))
    click.echo(clean_kernel(subdirectories, config.get(KEYS.SUB_PATTERNS),
                            root_files, slurm_outputs,
                            config.get(KEYS.DRYRUN)))



from . import __name__ as pkg_name
from ..utils.object_with_template import EnvironmentOfPackage, ObjectWithTemplateBase

env = EnvironmentOfPackage(pkg_name)


class ObjectWithTemplate(ObjectWithTemplateBase):
    pkg_env = env





from .base import ObjectWithTemplate
import random
from typing import Tuple
from warnings import warn


class Script(ObjectWithTemplate):
    template = None

    def add_task(self, task):
        raise NotImplementedError


class Task:
    def render():
        raise NotImplementedError


class TaskWithFileArg(Task):
    def __init__(self, filename):
        self.fn = filename


class GateSimulation(TaskWithFileArg):
    def render(self):
        return "Gate {}".format(self.fn)


class RootAnalysis(TaskWithFileArg):
    def __init__(self, root_filename, c_file_name):
        super().__init__(root_filename)
        self.c_file_name = c_file_name

    def render(self):
        return "root -q -b {} {}".format(self.fn, self.c_file_name)


class PygateAnalysis(TaskWithFileArg):
    def __init__(self, source=None, target=None, output=None, name=None, analysis_type=None):
        super().__init__(target)
        self.source = source
        self.target = target
        self.output = output
        self.name = name
        self.analysis_type = analysis_type

    def render(self):
        if self.analysis_type == 'script':
            fmt = "pygate analysis script --source {} --target {} --output {}"
            return fmt.format(self.source, self.fn, self.output)
        elif self.analysis_type == 'predefined':
            fmt = "pygate analysis predefined --source {} --output {} --name {}"
            return fmt.format(self.source, self.output, self.name)
        else:
            raise ValueError("Unknown analysis subcommand {}.".format(self.analysis_type))


class Clean(Task):
    def render(self):
        return "pygate clean --subdirectories"


class Merge(TaskWithFileArg):
    def __init__(self, filename, method):
        super().__init__(filename)
        self.method = method

    def render(self):
        return "pygate merge --target {} --method {}".format(self.fn, self.method)


class ScriptRun(Script):
    template = 'run'

    def __init__(self, work_directory, tasks: Tuple[Task]=(), geant4_version='8.0', shell='bash', is_need_source_env=False, partition='cpu'):
        """
        Parameters:

        - is_need_source_env: switch whether add line `souce ~/.{shell}rc` to shell script generated.
        """
        self.work_directory = work_directory
        self.geant4_version = geant4_version
        self.shell = shell
        self.is_need_source_env = is_need_source_env
        self.tasks = tasks
        self.partition = partition

    def add_task(self, task):
        return ScriptRun(self.work_directory, tuple(list(self.tasks) + [task]), self.geant4_version, self.shell, self.is_need_source_env)


class ScriptRunLocal(ScriptRun):
    template = 'run_local'

    def __init__(self, work_directory, tasks: Tuple[Task], geant4_version, shell='bash', is_need_source_env=False, local_work_directory=None, partition='cpu'):
        if local_work_directory is None:
            local_work_directory = '/tmp/pygate_temp_{}'.format(
                random.randint(0, 1e8))
        super().__init__(local_work_directory, tasks,
                         geant4_version, shell, is_need_source_env)
        warn(DeprecationWarning("No local run script is needed when using gluster."))
        self.server_work_directory = work_directory
        self.partition = partition

    def add_task(self, task):
        return ScriptRunLocal(self.work_directory, tuple(list(self.tasks) + [task]), self.geant4_version, self.shell, self.is_need_source_env, self.local_work_directory)


class ScriptPostRun(Script):
    template = 'post_run'

    def __init__(self, tasks: Tuple[Task]=(), geant4_version='8.0', shell='bash', is_need_source_env=False, partition='cpu'):
        self.tasks = tasks
        self.shell = shell
        self.geant4_version = geant4_version
        self.is_need_source_env = is_need_source_env
        self.partition = partition

    def add_task(self, task):
        return ScriptPostRun(tuple(list(self.tasks) + [task]), self.shell, self.is_need_source_env)


class ScriptMacTemplate(ObjectWithTemplate):
    template = 'make_mac'



from ..components.digitizer import *
from ..components.geometry import *


def ecat_digitizer(dtvolume,
                   rdr=None, blur=None,
                   thres=None, uph=None,
                   ddt=None, coin=None,
                   coin_delay=None,
                   coin_chain=None):
    ad = Adder()
    if rdr is None:
        rdr = Readout(depth=1)
    if blur is None:
        blur = Blurring(resolution=0.26, eor=511.0)
    if thres is None:
        thres = ThresHolder(value=250.0)
    if uph is None:
        uph = UpHolder(value=750.0)
    if ddt is None:
        ddt = DeadTime(volume=dtvolume, t=3000.0, mode='paralysable')
    singles = Singles([ad, rdr, blur, thres, uph, ddt])
    if coin is None:
        coin = CoincidenceSorter(window=10, offset=0)
    if coin_delay is None:
        coin_delay = CoincidenceSorter(name='delay', window=10, offset=500, is_define_name=True, is_explicit_insert=True)
    if coin_chain is None:
        coin_chain = CoincidencesChain(
            input1=coin, input2=coin_delay, name='finalCoinc', use_priority='True')
    digitizer = [singles, coin, coin_delay, coin_chain]
    return digitizer

def cylindricalPET_digitizer(rdr=None, blur=None,
                             thres=None, uph=None, 
                             coin=None, coin_delay=None):
    ad = Adder()
    if rdr is None:
        rdr = Readout(depth=1)
    if blur is None:
        blur = Blurring(resolution=0.26, eor=511.0)
    if thres is None:
        thres = ThresHolder(value=250.0)
    if uph is None:
        uph = UpHolder(value=650.0)
    singles = Singles([ad, rdr, blur, thres, uph])
    if coin is None:
        coin = CoincidenceSorter(window=10)
    if coin_delay is None:
        coin_delay = CoincidenceSorter(name='delay', window=10, offset=500,is_define_name=True, is_explicit_insert=False)
    digitizer = [singles, coin, coin_delay]
    return digitizer


def optical_digitizer(rdr = None):
    ad_op = AdderOptical()
    rdr = Readout(depth=1)
    singles = Singles([ad_op, rdr])
    digitizer = [singles,]
    return digitizer


def spect_digitizer(blur = None, spblur = None, thres = None, uph = None):
    raise NotImplementedError



"""
Warp of supported methods, set their working filesystem to current directory.
"""
from fs.osfs import OSFS


def make_config(target='.'):
    from .config_maker import ConfigMaker
    from .config import config as c
    with OSFS('.') as fs:
        ConfigMaker.make(fs, '.')


def init(config, method):
    from .initializer import Initializer
    with OSFS('.') as fs:
        initer = Initializer(fs, config)
        if method == 'pre' or method == 'all':
            initer.pre_sub()
        if method == 'sub' or method == 'all':
            initer.make_sub()


def submit(config):
    from .submitter import Submitter
    with OSFS('.') as fs:
        Submitter(fs, config).submit()


def merge(config):
    from .merger import Merger
    with OSFS('.') as fs:
        Merger(fs, config).merge()


def clean(config):
    from .cleaner import Cleaner
    with OSFS('.') as fs:
        Cleaner(fs, config).clean()



from dxl.fs import Directory, match_file
from .base import RoutineOnDirectory, OperationOnSubdirectories, Operation
from typing import Iterable


class KEYS:
    TO_REMOVE = 'remove'


class OpCleanSubdirectories(OperationOnSubdirectories):
    def apply(self, r: RoutineOnDirectory):
        result = self.dryrun(r)
        (self.subdirectories(r).map(lambda d: d.remove())
         .to_list().to_blocking().first())
        return result

    def dryrun(self, r: RoutineOnDirectory):
        return {KEYS.TO_REMOVE: (self.subdirectories(r).map(lambda d: d.path.s)
                                 .to_list().to_blocking().first())}


class OpCleanSource(Operation):
    def __init__(self, file_patterns: Iterable[str]):
        self.file_patterns = file_patterns

    def files(self, r: RoutineOnDirectory):
        return (r.directory.listdir_as_observable()
                .filter(match_file(self.file_patterns)))

    def apply(self, r: RoutineOnDirectory):
        result = self.dryrun(r)
        self.files(r).map(lambda f: f.remove()).to_list().to_blocking().first()
        return result

    def dryrun(self, r: RoutineOnDirectory):
        return {KEYS.TO_REMOVE: (self.files(r).map(lambda f: f.path.s)
                                 .to_list().to_blocking().first())}


class OpCleanFilesInAllDirectories(OperationOnSubdirectories):
    def __init__(self, file_patterns: Iterable[str],
                 subdirectory_patterns: Iterable[str]):
        super().__init__(subdirectory_patterns)
        self.file_patterns = file_patterns

    def files_on_directroy(self, d: Directory):
        return (d.listdir_as_observable()
                .filter(match_file(self.file_patterns)))

    def to_remove(self, r: RoutineOnDirectory):
        from dxl.fs import match_file
        files_root = self.files_on_directroy(r.directory)
        files_sub = (self.subdirectories(r)
                     .flat_map(lambda d: self.files_on_directroy(d)))
        files_all = files_root.merge(files_sub)
        return files_all

    def apply(self, r: RoutineOnDirectory):
        result = self.dryrun(r)
        (self.to_remove(r).map(lambda f: f.remove())
         .to_list().to_blocking().first())
        return result

    def dryrun(self, r: RoutineOnDirectory):
        return {KEYS.TO_REMOVE: (self.to_remove(r).map(lambda f: f.path.s)
                                 .to_list().to_blocking().first())}


def routine_clean(source_patterns: Iterable[str]=(), is_subdir: bool=False, dryrun: bool=False):
    """
    Helper function for clean in current directory.
    """
    ops = []
    if len(source_patterns) > 0:
        ops.append(OpCleanSource(source_patterns))
    if is_subdir:
        ops.append(OpCleanSubdirectories())
    return RoutineOnDirectory(Directory('.'), ops, dryrun)

# class Cleaner(RoutineOnDirectory):
#     def __init__(self, , is_clean_subdir, is_clean_source, split_name, dryrun=False):
#         ops = []
#         if is_clean_subdir:
#             ops.append(OpCleanSubdir())
#         if is_clean_source:
#             ops.append(OpCleanSource)
#         super().__init__(fs, tuple(ops), dryrun)
#         self.fs = fs
#         self.split_name = split_name

    # def clean(self):
    #     if self.c['sub_dirs']:
    #         self._clean_subs()
    #     if self.c['sources']:
    #         self._clean_sources()

    # def msg(self, path):
    #     from fs.path import normpath
    #     if self.c['verbose'] > 0:
    #         print('DELETE: {}'.format(normpath(self.fs.getsyspath(path))))

    # def _clean_subs(self):
    #     from dxpy.batch import DirectoriesFilter
    #     dirs = DirectoriesFilter(self.sub_dir).lst(self.fs)
    #     for d in dirs:
    #         self.msg(d)
    #         if not self.c['dryrun']:
    #             self.fs.removetree(d)

    # def _clean_sources(self):
    #     from dxpy.batch import FilesFilter
    #     files = FilesFilter(['*']).lst(self.fs)
    #     for f in files:
    #         if f in self.c['keep']:
    #             continue
    #         self.msg(f)
    #         if not self.c['dryrun']:
    #             self.fs.remove(f)



from ..components import ObjectWithTemplate
from .geometry import Geometry
from .misc import MaterialDatabaseLocal, Visualisation


class Simulation(ObjectWithTemplate):
    template = 'simulation'
    def __init__(self,
                 geometry,
                 physics,
                 digitizers,
                 source,
                 parameter,
                 material_database=None,
                 visualisation=None):
        self.geometry = geometry
        self.digitizers = digitizers
        self.physics = physics
        self.source = source
        self.parameter = parameter
        self.material_database = material_database or MaterialDatabaseLocal()
        self.visualisation = visualisation or Visualisation()



from .base import RoutineOnDirectory, OperationOnFile, OperationOnSubdirectories, OpeartionWithShellCall
from dxl.fs import Path, File, Directory
from ..utils.typing import JSONStr
import json

from rx import Observable
from typing import Iterable, Callable
from functools import partial


class OpMerge(OperationOnFile, OperationOnSubdirectories):
    """
    Operation merge files with same name in different subdirectories.
    """
    method = 'dryrun'

    def __init__(self, filename: str, patterns: Iterable[str]):
        OperationOnFile.__init__(self, filename)
        OperationOnSubdirectories.__init__(self, patterns)

    def sources(self, r: RoutineOnDirectory):
        return self.subdirectories(r).map(lambda d: d.attach_file(self.filename))

    def dryrun(self, r: RoutineOnDirectory) -> JSONStr:
        sources_system_path = (self.sources(r)
                               .map(lambda f: f.path.s)
                               .to_list().to_blocking().first())
        result = {
            'merge_method': self.method,
            'sources': sources_system_path,
        }
        result.update(OperationOnFile.dryrun(self, r))
        return result


# class RoutineMerge(RoutineOnDirectory):
#     def __init__(self, directory: Directory, operations: Iterable[Operation]=(), dryrun=False, verbose=0):
#         super().__init__(directory, operations, dryrun, verbose)

#         # class OpMergeMultipleFile(Operation):
#         #     def __init__(self, filename_patterns, subdir_pattern, op: OpMergeSingleFile):
#         #         self.fns = filename_patterns
#         #         self.sds = subdir_pattern
#         #         self.op = op

#         #     def get_sub_ops(self, r: RoutineOnDirectory):
#         #         return (r.list_matched_files(self.fns)
#         #                 .map(lambda f: self.op(f, self.sds)))

#         #     def apply(self, r: RoutineOnDirectory):
#         #         return (self.get_sub_ops(r)
#         #                 .map(lambda o: o.apply(r))
#         #                 .to_list().to_blocking.first())

#         #     def dryrun(self, r: RoutineOnDirectory):
#         #         return (self.get_sub_ops(r)
#         #                 .map(lambda o: o.dryrun(r))
#         #                 .to_list().to_blocking.first())
class OpMergeWithShellCall(OpMerge, OpeartionWithShellCall):
    def __init__(self, filename: str, patterns: Iterable[str]):
        OpMerge.__init__(self, filename, patterns)

    def apply(self, r: RoutineOnDirectory):
        result = OpMerge.apply(self, r)
        result.update(OpeartionWithShellCall.apply(self, r))
        return result

    def dryrun(self, r: RoutineOnDirectory):
        result = OpMerge.dryrun(self, r)
        result.update(OpeartionWithShellCall.dryrun(self, r))
        return result


class OpMergeHADD(OpMergeWithShellCall):
    method = 'hadd'

    def call_args(self, r: RoutineOnDirectory):
        target = self.target(r).path.s
        sources = (self.sources(r)
                   .map(lambda f: f.path.s)
                   .to_list().to_blocking().first())
        call_args = ['hadd', target] + sources
        call_args = [' '.join(call_args)]
        return call_args


class OpMergeCat(OpMergeWithShellCall):
    method = 'cat'

    def call_args(self, r: RoutineOnDirectory):
        target = self.target(r).path.s
        sources = (self.sources(r)
                   .map(lambda f: f.path.s)
                   .to_list().to_blocking().first())
        call_args = ['cat {} > {}'.format(' '.join(sorted(sources)), target)]
        # call_args = ['cat'] + sources + ['>', target]
        return call_args


class OpMergePandasConcatenate(OpMerge):
    method = 'pandas_concatenate'

    def apply(self, r: RoutineOnDirectory):
        result = self.dryrun(r)
        sources = self.sources(r).to_list().to_blocking().first()
        import pandas as pd
        tables = []
        for s in sources:
            p_ = s.path.s
            if p_.endswith('.csv'):
                tables.append(pd.read_csv(p_))
            elif p_.endswith('.h5'):
                tables.append(pd.read_hdf(p_))
            else:
                raise ValueError("Unknown file type {}".format(s.path.s))
        merged = pd.concat(tables, axis=0)
        target_path = self.target(r).path.s
        if target_path.endswith('.csv'):
            merged.to_csv(target_path)
        elif target_path.endswith('.h5'):
            merged.to_hdf(target_path, 'data')
        else:
            raise ValueError("Unknown file type {}".format(target_path))
        return result


class OpMergeSumBinary(OpMerge):
    method = 'sum_bin'


def hadd(work_directory: Directory, subdirectory_patterns: Iterable[str],
         source_filenames: Iterable[str], dryrun=False):
    """
    Helper function of fast create routine with only hadd task.
    """
    ops = [OpMergeHADD(s, subdirectory_patterns) for s in source_filenames]
    return RoutineOnDirectory(work_directory, ops, dryrun)

    # class OpMergeCat(Operation):
    #     method = 'cat'
    #     def apply(self, r:RoutineOnDirectory):

    #     with self.fs.open(target, 'w') as fout:
    #                 for f in sources:
    #                     with self.fs.open(f) as fin:
    #                         print(fin.read(), end='', file=fout

    # class OpMergeSum(Operation):
    #     def __init__(self, dtype, sources, target):
    #         super().__init__(sources, target)
    #         self.dtype = dtype

    # class OpMergeCopy(Operation):
    #     pass

    # class Merger:
    #     def __init__(self, fs, config):
    #         from .utils import sub_dir_filters
    #         self.fs = fs
    #         self.c = config['merge']
    #         self.sub_dirs = sub_dir_filters(config)

    #     def _path_of_file_in_sub_dirs(self, base_filename):
    #         from dxpy.batch import files_in_directories
    #         return files_in_directories(self.fs, self.sub_dirs, [base_filename])

    #     def msg(self, method, target, sources):
    #         if self.c.get('verbose') is not None and self.c['verbose'] > 0:
    #             print('MERGE.{md}.TARGET:'.format(md=method), target)
    #             print('MERGE.{md}.SOURCE:'.format(md=method), *sources, sep='\n')

    #     def _sum(self, task):
    #         import numpy as np
    #         if not task['method'].lower() == 'sum':
    #             return
    #         target = task['filename']
    #         sources = self._path_of_file_in_sub_dirs(target)
    #         self.msg('sum', target, sources)
    #         dtype = task['dtype']
    #         mapped = {
    #             'float32': np.float32,
    #             'uint16': np.uint16,
    #             'uint8': np.uint8,
    #         }
    #         dtype = mapped[dtype]
    #         result = None
    #         for f in sources:
    #             with self.fs.open(f, 'rb') as fin:
    #                 data = np.fromstring(fin.read(), dtype=dtype)
    #                 if result is None:
    #                     result = data
    #                 else:
    #                     result += data
    #         with self.fs.open(target, 'wb') as fout:
    #             fout.write(result.tostring())

    #     def _copy(self, task):
    #         import numpy as np
    #         if not task['method'].lower() == 'sum':
    #             return
    #         target = task['filename']
    #         sources = self._path_of_file_in_sub_dirs(target)
    #         if len(sources):
    #             raise ValueError("Sources not found")
    #         from fs.copy import copy_file
    #         self.msg('copy', target, sources[0])
    #         copy_file(self.fs, sources[0], self.fs, target)

    #     def merge(self):
    #         supported_methods = [self._hadd, self._cat, self._sum]
    #         for t in self.c['tasks']:
    #             for func in supported_methods:
    #                 func(t)



from ..base import ObjectWithTemplate
from .volume import Volume

class Surface(ObjectWithTemplate):

    template = 'geometry/surface'
    surface_type = None

    def __init__(self, name, base: Volume, insert: Volume):
        self.name = name
        self.base = base
        self.insert = insert


class SurfacePerfectAPD(Surface):
    surface_type = 'perfect_apd'

class SurfaceRoughTeflonWrapped(Surface):
    surface_type  = 'rough_teflon_wrapped'


from .base import ObjectWithTemplate


class Insertable(ObjectWithTemplate):
    template = 'digitizer/insertable'
    insertable_type = None

    def __init__(self, name=None, is_define_name=False, is_explicit_insert = True):
        self.name = name
        self.base = None
        self.is_define_name = is_define_name
        self.is_explict_insert = is_explicit_insert


class Singles(Insertable):
    template = 'digitizer/singles/singles'

    def __init__(self, plugins=None, name='Singles', is_define_name=False, is_explicit_insert=False):
        super().__init__(name, is_define_name, is_explicit_insert)
        self.plugins = plugins
        for p in self.plugins:
            p.base = self


class Adder(Insertable):
    insertable_type = 'adder'

    def __init__(self, name='adder', is_define_name=False):
        super().__init__(name, is_define_name)

class AdderCompton(Insertable):
    insertable_type = 'adderCompton'
    def __init__(self, name='adderCompton', is_define_name=False):
        super().__init__(name, is_define_name)

class AdderOptical(Insertable):
    insertable_type = 'opticaladder'
    def __init__(self, name='opticaladder', is_define_name = False):
        super().__init__(name,is_define_name)


class Readout(Insertable):
    template = 'digitizer/singles/readout'
    insertable_type = 'readout'

    def __init__(self, policy=None, depth=1,
                 name='readout', is_define_name=False):
        super().__init__(name, is_define_name)
        self.policy = policy
        self.depth = depth


class Blurring(Insertable):
    template = 'digitizer/singles/blurring'
    insertable_type = 'blurring'

    def __init__(self, law=None, resolution=0.15, eor=511, slope=None,
                 name='blurring', is_define_name=False):
        super().__init__(name, is_define_name)
        self.law = law
        self.resolution = resolution
        self.eor = eor
        self.slope = slope


class Holder(Insertable):
    template = 'digitizer/singles/holder'
    holder_name = None
    insertable_type = None

    def __init__(self, value, name=None, is_define_name=False):
        super().__init__(name, is_define_name)
        self.value = value


class ThresHolder(Holder):
    holder_name = 'Threshold'
    insertable_type = 'thresholder'

    def __init__(self, value, name='thresholder'):
        super().__init__(value, name)


class UpHolder(Holder):
    holder_name = 'Uphold'
    insertable_type = 'upholder'

    def __init__(self, value, name='upholder'):
        super().__init__(value, name)


class TimeResolution(Insertable):
    template = 'digitizer/singles/time_resolution'
    insertable_type = 'timeResolution'

    def __init__(self, resolution, name=None, is_define_name=False):
        super().__init__(name, is_define_name)
        self.resolution = resolution


class WithBuffer(Insertable):
    template = 'digitizer/singles/buffer'

    def __init__(self, size=None, mode=None, name=None, is_define_name=False):
        super().__init__(name, is_define_name)
        self.size = size
        self.mode = mode


class MemoryBuffer(WithBuffer):
    template = 'digitizer/singles/memory_buffer'
    insertable_type = 'buffer'

    def __init__(self, read_freq=None, size=None, mode=None, name=None, is_define_name=False):
        super().__init__(name, size, mode, name, is_define_name)
        self.read_freq = read_freq


class DeadTime(WithBuffer):
    template = 'digitizer/singles/dead_time'
    insertable_type = 'deadtime'

    def __init__(self, volume, t, mode=None, buffer_size=None, buffer_mode=None, name='deadtime', is_define_name=False):
        super().__init__(buffer_size, buffer_mode, name, is_define_name)
        self.volume = volume
        self.t = t
        self.mode = mode

class DeadTimeMulti(WithBuffer):
    template = 'digitizer/singles/dead_time_multi'
    insertable_type = 'deadtimeMulti'
    def __init__(self, volume, t, mode=None, buffer_size=None, buffer_mode=None, name='deadtimeMulti', is_define_name=False):
        super().__init__(buffer_size, buffer_mode, name, is_define_name)
        self.volume = volume
        self.t = t
        self.mode = mode


class SinglesChain(Singles):
    def __init__(self, name, is_define_name=True):
        super().__init__(name, is_define_name)


class CoincidenceSorter(Insertable):
    template = 'digitizer/coincidence_sorter'
    insertable_type = 'coincidenceSorter'

    def __init__(self, input_=None, window=None, offset=None, name='Coincidences', is_define_name=False, is_explicit_insert = False):
        super().__init__(name, is_define_name, is_explicit_insert)
        self.input_ = input_
        self.window = window
        self.offset = offset


class CoincidencesChain(Insertable):
    template = 'digitizer/coincidence_chain'
    insertable_type = 'coincidenceChain'
    def __init__(self, input1, input2, name, plugins = None, use_priority=None, conserve_all_event = None, is_define_name=True):
        super().__init__(name, is_define_name)
        self.input1 = input1
        self.input2 = input2
        self.use_priority = use_priority
        self.conserve_all_event = conserve_all_event
        self.plugins = plugins or ()
        for p in self.plugins:
            p.base = self



class KEYS:
    SUBMIT = 'submit'
    DRYRUN = 'dryrun'
    SUB_PREFIX = 'sub_prefix'
    SUB_PATTERNS = 'sub_patterns'
    SUB_FORMAT = 'sub_format'
    CLEAN = 'clean'
    ANALYSIS = 'analysis'
    INIT = 'init'
    NB_SPLIT = 'nb_split'


class SUBMIT_KEYS:
    BROADCAST = 'broadcast'
    SINGLE = 'single'


class CLEAN_KEYS:
    IS_SUBDIRECTORIES = 'is_subdirectories'
    ROOT_FILES = 'root_files'
    IS_SLURM_OUTPUTS = 'is_slurm_outputs'


class ANALYSIS_KEYS:
    SOURCE = 'source'
    TARGET = 'target'
    NAME = 'name'
    OUTPUT = 'output'
    ANALYSIS_TYPE = 'subcommand'


class INIT_KEYS:

    BROADCAST = 'broadcast'

    class BROADCAST_KEYS:
        TARGETS = 'targets'
        NO_EXT = 'no_ext'

    EXTERNAL = 'external'

    class EXTERNAL_KEYS:
        SOURCE = 'source'
        TARGET = 'target'

    MAC = 'mac'

    class MAC_KEYS:
        SIMULATION_DEMO = 'simulation_demo'

    SHELL = 'shell'

    class SHELL_KEYS:
        RUN = 'run'
        POST_RUN = 'post_run'
        TASK = 'task'
        TASK_NAME = 'task_name'
        TARGET = 'target'
        GATE_SIMULATION = 'gate_simulation'
        ROOT_ANALYSIS = 'root_analysis'
        ROOT_C_FILE = 'root_c_file'
        MERGE = 'merge'
        SOURCE = 'source'
        METHOD = 'method'
        GATE_VERSION = 'gate_version'
        SHELL_TYPE = 'shell_type'
        PYGATE_ANALYSIS = 'pygate_analysis'
        PARTITION = 'partition'

    PHANTOM = 'phantom'

    class PHANTOM_KEYS:
        pass


shell_run_config = {
    INIT_KEYS.SHELL_KEYS.GATE_VERSION: '8.0',
    INIT_KEYS.SHELL_KEYS.SHELL_TYPE: 'bash',
    INIT_KEYS.SHELL_KEYS.TASK: [
        {INIT_KEYS.SHELL_KEYS.TASK_NAME: INIT_KEYS.SHELL_KEYS.GATE_SIMULATION,
         INIT_KEYS.SHELL_KEYS.TARGET: 'main.mac'},
    ],
    INIT_KEYS.SHELL_KEYS.TARGET: 'run.sh',
    INIT_KEYS.SHELL_KEYS.PARTITION: 'cpu',
}

shell_post_run_config = {
    INIT_KEYS.SHELL_KEYS.GATE_VERSION: '8.0',
    INIT_KEYS.SHELL_KEYS.TASK: [
        {INIT_KEYS.SHELL_KEYS.TASK_NAME:  INIT_KEYS.SHELL_KEYS.MERGE,
         INIT_KEYS.SHELL_KEYS.TARGET: 'result.root',
         INIT_KEYS.SHELL_KEYS.METHOD: 'hadd'},
        {INIT_KEYS.SHELL_KEYS.TASK_NAME: INIT_KEYS.SHELL_KEYS.ROOT_ANALYSIS,
         INIT_KEYS.SHELL_KEYS.TARGET: 'result.root',
         INIT_KEYS.SHELL_KEYS.ROOT_C_FILE: 'Hits2CSV.C'},
        {INIT_KEYS.SHELL_KEYS.TASK_NAME: INIT_KEYS.SHELL_KEYS.PYGATE_ANALYSIS,
         ANALYSIS_KEYS.SOURCE: 'optical.csv',
         ANALYSIS_KEYS.TARGET: 'optical.h5',
         ANALYSIS_KEYS.ANALYSIS_TYPE: 'gamma_edep'}
    ],
    INIT_KEYS.SHELL_KEYS.TARGET: 'post.sh',
    INIT_KEYS.SHELL_KEYS.SHELL_TYPE: 'bash',
    INIT_KEYS.SHELL_KEYS.PARTITION: 'cpu',
}

init_config = {
    INIT_KEYS.EXTERNAL: [
        {INIT_KEYS.EXTERNAL_KEYS.SOURCE: '/mnt/gluster_NoGPU/share/pygate/GateMaterials.db'},
        {INIT_KEYS.EXTERNAL_KEYS.SOURCE: '/mnt/gluster_NoGPU/share/pygate/Materials.xml'},
        {INIT_KEYS.EXTERNAL_KEYS.SOURCE: '/mnt/gluster_NoGPU/share/pygate/Surfaces.xml'},
        {INIT_KEYS.EXTERNAL_KEYS.SOURCE: '/mnt/gluster_NoGPU/share/pygate/Hits2CSV.C'},
    ],
    INIT_KEYS.SHELL: {
        INIT_KEYS.SHELL_KEYS.RUN: shell_run_config,
        INIT_KEYS.SHELL_KEYS.POST_RUN: shell_post_run_config,
    },
    INIT_KEYS.BROADCAST: {
        INIT_KEYS.BROADCAST_KEYS.TARGETS: ['main.mac'],
        INIT_KEYS.BROADCAST_KEYS.NO_EXT: False,
    },
}


config = {
    KEYS.SUBMIT: {
        SUBMIT_KEYS.BROADCAST: ['run.sh'],
        SUBMIT_KEYS.SINGLE: ['post.sh'],
    },
    KEYS.DRYRUN: False,
    KEYS.SUB_PREFIX: 'sub',
    KEYS.CLEAN: {
        CLEAN_KEYS.IS_SUBDIRECTORIES: True,
        CLEAN_KEYS.IS_SLURM_OUTPUTS: False,
        CLEAN_KEYS.ROOT_FILES: [],
    },
    KEYS.INIT: init_config,
    KEYS.NB_SPLIT: 10,
    KEYS.ANALYSIS: {
        ANALYSIS_KEYS.SOURCE: 'result.csv',
        ANALYSIS_KEYS.TARGET: 'result.h5',
        ANALYSIS_KEYS.ANALYSIS_TYPE: 'gamma_edep'
    }



}



from ..base import ObjectWithTemplate
from .geometry import Volume
from typing import Tuple


class Phantom(ObjectWithTemplate):
    template = 'geometry/phantom'

    def __init__(self, sensitive_detectors: Tuple[Volume]=()):
        self.sds = sensitive_detectors



def get_scripts_path():
    import pygate
    from dxpy.filesystem import Path
    p = Path(pygate.__file__)
    return Path(p.father) / 'scripts'


def load_script(name):
    script_file = (get_scripts_path() / name).abs
    with open(script_file) as fin:
        return fin.read()


def sub_dir_filters(config):
    return ['{name}*'.format(name=config['split']['name'])]



import unittest
from ruamel.yaml import YAML
from fs.memoryfs import MemoryFS
from fs.osfs import OSFS
from dxpy.filesystem import Path
import jinja2
from pygate import shell


# class TestShellScript(unittest.TestCase):
#     def test_shell_bash(self):
#         from fs.osfs import OSFS
#         result = shell.make(OSFS('/tmp'), '/sub.1', 'map', [{'method':'Gate', 'filename': 'main.mac'}], shell='bash')
#         from pygate.utils import load_script
#         expect = load_script('map_bash_sample.sh')
#         self.assertEqual(result, expect)
        




from .results import *
from dxl.fs import File


def gamma_energy_deposit_distribution(csv_filename, h5_filename):
    events = (CSVFile(csv_filename).load()
              .split_by(ColumnNames.Particle)
              .select(ParticleID.Gamma.value)
              .split_by(ColumnNames.Event).drop_keys()
              .map(lambda e: e.to_event()))

    # incident_directions = (events.map(lambda e: e.incident_direction())
    #    .map(lambda v: v.to_list()).to_list())
    # source_postions = (events.map(lambda e: e.source_postions())
    #    .map(lambda v: v.to_list()).to_list())
    # deposits = events.
    eps = events.map(lambda e: e.energy_deposit_list()).flatten()
    poss = eps.map(lambda ep: ep.position).map(lambda p: p.to_list())
    engs = eps.map(lambda ep: ep.energy)
    rezip = poss.zip(engs).map(lambda t: tuple(list(t[0]) + [t[1]]))
    df = pd.DataFrame(data=np.array(rezip.d), columns=[
                      'x', 'y', 'z', 'energy'])
    df.to_hdf(h5_filename, 'gamma_edep')
    return h5_filename



import click
from .main import pygate
from ..conf import config, KEYS
from ..utils.syscall import shell_call


def analysis_kernel(source, target, analysis_type, dryrun):
    from pygate.routine.analysis import OperationAnalysis
    from pygate.routine.base import RoutineOnDirectory
    from dxl.fs import Directory
    d = Directory('.')
    o = OperationAnalysis(source, target, analysis_type)
    r = RoutineOnDirectory(d, [o])
    r.work()
    return r.echo()


@pygate.group()
def analysis():
    pass


@analysis.command()
@click.option('--name', '-n', help="Predefined analysis type name.")
@click.option('--source', '-s', help="Analysis source data filename.")
@click.option('--output', '-o', help="Analysis target data filename.")
def predefined(name, source, output):
    click.echo(analysis_kernel(source, output, name, config.get(KEYS.DRYRUN)))


@analysis.command()
@click.option('--target', '-t', help="Analysis .py filename.")
@click.option('--source', '-s', help="Analysis source data filename.")
@click.option('--output', '-o', help="Output filename.")
def script(target, source, output):
    shell_call('python {} --source {} --output {}'.format(target, source, output))



from ...base import ObjectWithTemplate
from ..volume import Volume
from .system import System
from typing import Tuple


class Camera(ObjectWithTemplate):
    template = 'geometry/camera'

    def __init__(self,
                 system: System,
                 sensitive_detectors: Tuple[Volume]=()):
        self.system = system
        self.sds = sensitive_detectors



"""
Initializer provide the following features:
    1. copy sources from template directory
    2. generate mac files
    3. generate sub directory structure
    4. copy sources/macs to sub directories
    5. generate map shell script to sub directories
    6. generate merge shell script in root directory

Grouped by two methods:
    1. pre_sub 1,2
    2. make_sub 3,4,5,6     
"""


class Initializer:
    def __init__(self, fs, config):
        self.fs = fs
        self.c = config['init']
        self.c_full = config

    def _copy_sources_from_template(self):
        from fs.path import relativefrom
        from fs.copy import copy_file
        from fs.osfs import OSFS
        path_source = self.c['source']['directory']
        if not path_source.startswith('/'):
            path_source = self.fs.getsyspath(path_source)
        with OSFS('/') as fs:
            if not fs.exists(path_source):
                raise ValueError(
                    "Source path {} not exists.".format(path_source))
        with OSFS(path_source) as fs_sor:

            for f in self.c['source']['filenames']:
                copy_file(fs_sor, f, self.fs, f)

    def _make_mac(self):
        from .template import MacMaker
        MacMaker.make_mac(self.c['generate']['config'])

    def _make_phantom(self):
        from .phantom import PhantomBinFileMaker
        PhantomBinFileMaker(self.fs, self.c['phantom']).make()

    def _sub_dir_name(self, i):
        return '{name}.{did}'.format(name=self.c_full['split']['name'], did=i)

    @property
    def _nb_sub_dirs(self):
        return self.c_full['split']['nb_split']

    def _make_subdirs(self):
        for i in range(self._nb_sub_dirs):
            self.fs.makedir(self._sub_dir_name(i), recreate=True)

    def _make_map_shell_scripts(self):
        from .shell import ShellScriptMaker
        ssm = ShellScriptMaker(self.fs, self.c_full)
        ssm.make()

    def _copy_sources_to_subdirs(self):
        from dxpy import batch
        from fs.copy import copy_file
        all_sources = batch.FilesFilter(['*']).lst(self.fs)
        for i in range(self._nb_sub_dirs):
            with self.fs.opendir(self._sub_dir_name(i)) as d:
                for f in all_sources:
                    copy_file(self.fs, f, d, f)

    def pre_sub(self):
        if 'source' in self.c:
            self._copy_sources_from_template()
        if 'generate' in self.c:
            self._make_mac()
        if 'phantom' in self.c:
            self._make_phantom()

    def make_sub(self):
        self._make_subdirs()
        self._copy_sources_to_subdirs()
        self._make_map_shell_scripts()



class Merger:
    def __init__(self, fs, config):
        from .utils import sub_dir_filters
        self.fs = fs
        self.c = config['merge']
        self.sub_dirs = sub_dir_filters(config)

    def _path_of_file_in_sub_dirs(self, base_filename):
        from dxpy.batch import files_in_directories
        return files_in_directories(self.fs, self.sub_dirs, [base_filename])

    def msg(self, method, target, sources):
        if self.c.get('verbose') is not None and self.c['verbose'] > 0:
            print('MERGE.{md}.TARGET:'.format(md=method), target)
            print('MERGE.{md}.SOURCE:'.format(md=method), *sources, sep='\n')

    def _hadd(self, task):
        import subprocess
        import sys
        if not task['method'].lower() == 'hadd':
            return
        filename = task['filename']
        sub_filenames = self._path_of_file_in_sub_dirs(filename)
        path_target = self.fs.getsyspath(filename)
        path_filenames = [self.fs.getsyspath(f)
                          for f in sub_filenames if self.fs.exists(f)]
        call_args = ['hadd', path_target] + path_filenames
        self.msg('HADD', filename, sub_filenames)
        if not self.c['dryrun']:
            with subprocess.Popen(call_args,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE) as p:
                sys.stdout.write(p.stdout.read().decode())
                sys.stderr.write(p.stderr.read().decode())

    def _cat(self, task):
        if not task['method'].lower() == 'cat':
            return
        target = task['filename']
        sources = self._path_of_file_in_sub_dirs(target)
        self.msg('cat', target, sources)
        if not self.c['dryrun']:
            with self.fs.open(target, 'w') as fout:
                for f in sources:
                    with self.fs.open(f) as fin:
                        print(fin.read(), end='', file=fout)

    def _sum(self, task):
        import numpy as np
        if not task['method'].lower() == 'sum':
            return
        target = task['filename']
        sources = self._path_of_file_in_sub_dirs(target)
        self.msg('sum', target, sources)
        dtype = task['dtype']
        mapped = {
            'float32': np.float32,
            'uint16': np.uint16,
            'uint8': np.uint8,
        }
        dtype = mapped[dtype]
        result = None
        for f in sources:
            with self.fs.open(f, 'rb') as fin:
                data = np.fromstring(fin.read(), dtype=dtype)
                if result is None:
                    result = data
                else:
                    result += data
        with self.fs.open(target, 'wb') as fout:
            fout.write(result.tostring())

    def _copy(self, task):
        import numpy as np
        if not task['method'].lower() == 'sum':
            return
        target = task['filename']
        sources = self._path_of_file_in_sub_dirs(target)
        if len(sources):
            raise ValueError("Sources not found")
        from fs.copy import copy_file
        self.msg('copy', target, sources[0])
        copy_file(self.fs, sources[0], self.fs, target)

    def merge(self):
        supported_methods = [self._hadd, self._cat, self._sum]
        for t in self.c['tasks']:
            for func in supported_methods:
                func(t)



from ..components.geometry import *


def cylindricalPET(world: Volume, cylinder=None, rrh=None,
                   head=None, rcb=None, block=None, rcc=None,
                   crystal=None, lso=None, bgo=None):
    if cylinder is None:
        cylinder = Cylinder('cylindricalPET', rmax=52.0, rmin=39.9, height=40.2, material='Air',
                            mother=world, position=Vec3(0.0, 0.0, 0.0, 'cm'), unit='cm')
    if rrh is None:
        rrh = RepeaterRing(number=4)
    if head is None:
        head = Box('head', Vec3(8, 32, 40, 'cm'), 'Air',
                   cylinder, Vec3(44, 0, 0, 'cm'), repeaters=[rrh])
    if rcb is None:
        rcb = RepeaterCubic(scale=Vec3(1, 20, 20, ''),
                            repeat_vector=Vec3(0.0, 1.6, 2.0, 'cm'))
    if block is None:
        block = Box('block', Vec3(30, 16, 20, 'mm'),
                    'Air', head, repeaters=[rcb])
    if rcc is None:
        rcc = RepeaterCubic(scale=Vec3(1, 5, 5),
                            repeat_vector=Vec3(0.0, 3.2, 4.0, 'cm'))
    if crystal is None:
        crystal = Box('crystal', Vec3(30, 3, 3.8, 'mm'),
                      'Air', block, repeaters=[rcc])
    if lso is None:
        lso = Box('LSO', Vec3(15, 3.0, 3.8), 'LSO',
                  crystal, Vec3(-0.75, 0.0, 0.0, 'cm'))
    if bgo is None:
        bgo = Box('BGO', Vec3(15, 3.0, 3.8), 'BGO',
                  crystal, Vec3(0.75, 0.0, 0.0, 'cm'))

    sys = CylindricalPET(
        head, block, crystal=crystal, layer0=lso, layer1=bgo)
    sen_list = [crystal, ]
    cam = Camera(sys, sen_list)
    return cam


def ecat(world: Volume, cylinder=None, rlb=None, rrb=None,
         block=None, rcc=None, crystal=None):
    if cylinder is None:
        cylinder = Cylinder('ecat', rmax=44.2, rmin=41.2, height=15.52, material='Air',
                            mother=world, unit='cm')
    if rlb is None:
        rlb = RepeaterLinear(number=4, repeat_vector=Vec3(0, 0, 38.8, 'mm'))
    if rrb is None:
        rrb = RepeaterRing(number=72)
    if block is None:
        block = Box('block', size=Vec3(30, 35.8594, 38.7, 'mm'), position=Vec3(427.0, 0.0, 0.0, 'mm'),
                    material='Air', mother=cylinder, repeaters=[rlb, rrb])
    if rcc is None:
        rcc = RepeaterCubic(scale=Vec3(1, 8, 8, ''),
                            repeat_vector=Vec3(0, 4.4942, 4.85, 'mm'))
    if crystal is None:
        crystal = Box('crystal', size=Vec3(30.0, 4.4, 4.75, 'mm'),
                      material='BGO', mother=block, repeaters=[rcc])

    sys = Ecat(block, crystal)
    sen_list = [crystal, ]
    cam = Camera(sys, sen_list)
    return cam


def opticalsystem(world: Volume, box=None, crystal=None,
                  rcp=None, pixel=None):
    if box is None:
        box = Box('OpticalSystem', size=Vec3(
            5.0, 5.0, 5.0, 'cm'), material='Air', mother=world)
    if crystal is None:
        crystal = Box('crystal', Vec3(30, 30, 10, 'mm'), position=Vec3(
            0.0, 0.0, 5.0, 'mm'), material='LYSO', mother=box)
    if rcp is None:
        rcp = RepeaterCubic(
            Vec3(10, 10, 1), repeat_vector=Vec3(3.0, 3.0, 0.0, 'mm'))
    if pixel is None:
        pixel = Box('pixel', Vec3(3.0, 3.0, 1.0), position=Vec3(
            0, 0, -0.5, 'mm'), material='G4_SILICON_DIOXIDE', mother=crystal, repeaters=[rcp])
    sys = OpticalSystem(crystal, pixel)
    sen_list = [crystal, ]
    cam = Camera(sys, sen_list)
    return cam


from typing import TypeVar, Callable

VolumeMaker = Callable[[Volume], Volume]
VolumeLike = TypeVar('VolumeLike', Volume, VolumeMaker)


def optical_gamma(world: Volume, crystal: VolumeLike):
    box = Box('OpticalSystem', size=world.size, material='Air', mother=world)
    if not isinstance(crystal, Volume):
        crystal = crystal(box)
    if crystal is None:
        crystal = Box('crystal', Vec3(30, 30, 10, 'mm'), position=Vec3(
            0.0, 0.0, 5.0, 'mm'), material='LYSO', mother=box)
    return Camera(OpticalSystem(crystal), (crystal,))


def multipatchPET(world: Volume):
    world = Box('world', Vec3(50, 50, 50, 'cm'))

    box = Box('multiPatchPET', size=Vec3(
        40.0, 40.0, 40.0, 'cm'), material='Air', mother=world)
    container = Sphere('container', rmax=17, rmin=15, phi_start=0, delta_phi=360, theta_start=0, delta_theta=180,
                       material='Air', mother=box)
    patch_list = Patch('patch', patch_file='patch1.pat',
                       material='LYSO', mother=container)
    # to be completed
    sys = MultiPatchPET(container, patch_list)
    sen_list = [patch_list]
    cam = Camera(sys, sen_list)
    return cam


def optical_surfaces(cam: Camera):
    surface1 = SurfaceRoughTeflonWrapped(
        'surface1', cam.system.levels['crystal'], cam.system.levels['crystal'].mother)
    surface2 = SurfaceRoughTeflonWrapped(
        'surface2', cam.system.levels['crystal'].mother, cam.system.levels['crystal'])
    Detection1 = SurfacePerfectAPD(
        'Detection1', cam.system.levels['crystal'], cam.system.levels['pixel'])
    Detection2 = SurfacePerfectAPD(
        'Detection2', cam.system.levels['pixel'], cam.system.levels['crystal'])
    surfaces = [surface1, surface2, Detection1, Detection2]
    return surfaces


# class Camera:
#     def __init__(self, name, world=None, detector=None):
#         self.name = name
#         self.world = None
#         self.detector = detector
#         self.geo_list = []
#         self.crystalSD_list = []
#         self.attach_list = []

#     def add_geo(self, item):
#         self.geo_list.append(item)

#     def add_crystalSD(self, item):
#         self.crystalSD_list.append(item)

#     # called after the geometry is defined in deritives.
#     def set_world(self, world):
#         if world is None:
#             maxsize = self.detector.getDiameter()
#             # set the world size to 5 time of the diamter of the detector if it is not given.
#             worldsize = geometry.Vec3(5*maxsize, 5*maxsize, 5*maxsize)
#             self.world = geometry.Box(name='world', size=worldsize)
#         else:
#             self.world = world

#     def composite(self):
#         self.world.add_child(self.detector)



"""
Initialization CLI support.

Note many configurations is done by config file.
"""
import click
from .main import pygate
from ..conf import config, KEYS, INIT_KEYS
import json
import yaml
from dxl.fs import File, Directory

from ..utils.syscall import shell_call

@pygate.group()
def init():
    pass


@pygate.group()
def generate():
    pass

from .mac_generate import mac

@generate.command()
@click.option('--filename', '-f', default='make_mac.py')
def mac_template(filename):
    d = Directory('.')
    from pygate.routine.initialize import OpGenerateMacTemplate, RoutineOnDirectory
    o = OpGenerateMacTemplate(filename)
    r = RoutineOnDirectory(d, [o])
    r.work()
    # click.echo(json.dumps(r.echo(), indent=4, separators=(',', ': ')))


def shell_task_list(tasks):
    from pygate.scripts.shell import ScriptRun, GateSimulation, RootAnalysis, Merge, PygateAnalysis
    from ..conf import ANALYSIS_KEYS
    result = []
    SKS = INIT_KEYS.SHELL_KEYS
    for t in tasks:
        name = t[SKS.TASK_NAME]
        if name == SKS.GATE_SIMULATION:
            result.append(GateSimulation(t[SKS.TARGET]))
        elif name == SKS.MERGE:
            result.append(Merge(t[SKS.TARGET], t[SKS.METHOD]))
        elif name == SKS.ROOT_ANALYSIS:
            result.append(RootAnalysis(t[SKS.TARGET], t[SKS.ROOT_C_FILE]))
        elif name == SKS.PYGATE_ANALYSIS:
            result.append(PygateAnalysis(t.get(ANALYSIS_KEYS.SOURCE),
                                         t.get(ANALYSIS_KEYS.TARGET),
                                         t.get(ANALYSIS_KEYS.OUTPUT),
                                         t.get(ANALYSIS_KEYS.NAME),
                                         t.get(ANALYSIS_KEYS.ANALYSIS_TYPE)))
        else:
            raise ValueError("Unknown task name {}.".format(name))
    return result


def shell_run(filename, tasks, gate_version, shell, partition):
    from pygate.scripts.shell import ScriptRun, GateSimulation, RootAnalysis
    SKS = INIT_KEYS.SHELL_KEYS
    shell_script = ScriptRun(Directory('.').system_path(), shell_task_list(tasks),
                             gate_version, shell, partition=partition)
    with open(filename, 'w') as fout:
        print(shell_script.render(), file=fout)


def shell_post_run(filename, tasks, gate_version, shell, partition):
    from pygate.scripts.shell import ScriptPostRun, Merge, RootAnalysis
    SKS = INIT_KEYS.SHELL_KEYS
    shell_script = ScriptPostRun(
        shell_task_list(tasks), gate_version, shell, partition=partition)
    with open(filename, 'w') as fout:
        print(shell_script.render(), file=fout)


@generate.command()
def shell():
    """
    Generate shell script, pre run or post run.
    """
    SKS = INIT_KEYS.SHELL_KEYS
    shellc = config.get(KEYS.INIT, {}).get(INIT_KEYS.SHELL)
    src = shellc.get(SKS.RUN)
    shell_run(src[SKS.TARGET], src[SKS.TASK],
              src[SKS.GATE_VERSION], src[SKS.SHELL_TYPE],
              src[SKS.PARTITION])
    sprc = shellc.get(SKS.POST_RUN)
    shell_post_run(sprc[SKS.TARGET], sprc[SKS.TASK], src[SKS.GATE_VERSION],
                   sprc[SKS.SHELL_TYPE],
                   sprc[SKS.PARTITION])


@generate.command()
@click.option('--target', '-t', help='Config file name.')
@click.option('--format', '-f', help='Format of config file, json or yml')
def cfg(target, format):
    """
    Generate initial config file.
    """
    from dxl.fs import Path
    if target is None and format is None:
        target = 'pygate.yml'
    if format is None:
        format = Path(target).e
    if target is None:
        target = 'pygate.{}'.format(format)
    if format.startswith('.'):
        format = format[1:]
    with open(target, 'w') as fout:
        if format.lower() == 'yml':
            yaml.dump(config, fout)
        elif format.lower() == 'json':
            json.dump(config, fout, indent=4, sort_keys=True)


@init.command()
@click.option('--nb-split', '-n', type=int, help='Number of subdirectories.')
@click.option('--sub-format', '-f', help='Subdirectories format str.')
def subdir(nb_split, sub_format):
    from pygate.routine.initialize import OpSubdirectoriesMaker, RoutineOnDirectory
    if nb_split is None:
        nb_split = config.get(KEYS.NB_SPLIT)
    if sub_format is None:
        sub_format = config.get(KEYS.SUB_FORMAT)
    d = Directory('.')
    op = OpSubdirectoriesMaker(nb_split, sub_format)
    r = RoutineOnDirectory(d, [op], config.get(KEYS.DRYRUN))
    r.work()
    click.echo(r.echo())


def broadcast_kernel(files, subdirectory_patterns, dryrun):
    from pygate.routine.initialize import OpAddToBroadcastFile, OpBroadcastFile, RoutineOnDirectory
    d = Directory('.')
    ops = [OpAddToBroadcastFile(f) for f in files]
    ops.append(OpBroadcastFile(subdirectory_patterns))
    r = RoutineOnDirectory(d, ops, dryrun)
    r.work()
    return r.echo()


@init.command()
@click.option('--target', '-t', type=int, help='Files to broadcast to subdirectories.', multiple=True)
@click.option('--no-ext', '-e', help='Include all external files.', is_flag=True)
def bcast(target, no_ext):
    if no_ext is None:
        no_ext = (config.get(KEYS.INIT)
                  .get(INIT_KEYS.BROADCAST, {})
                  .get(INIT_KEYS.BROADCAST_KEYS.ADD_EXT))
    if len(target) == 0:
        target = (config.get(KEYS.INIT, {})
                  .get(INIT_KEYS.BROADCAST, {})
                  .get(INIT_KEYS.BROADCAST_KEYS.TARGETS, ()))
    files = list(target)
    fmt = 'pygate init bcast called with: target: {} no-ext: {}.'
    click.echo(fmt.format(target, no_ext))
    if not no_ext:
        files = list(files) + [e[1] for e in external_to_copy()]
    click.echo(broadcast_kernel(files, config.get(KEYS.SUB_PATTERNS),
                                config.get(KEYS.DRYRUN)))


def external_to_copy():
    from dxl.fs import Path
    inic = config.get(KEYS.INIT)
    tasks = inic.get(INIT_KEYS.EXTERNAL)
    if tasks is None:
        return ()
    result = []
    for t in tasks:
        source = t[INIT_KEYS.EXTERNAL_KEYS.SOURCE]
        target = t.get(INIT_KEYS.EXTERNAL_KEYS.TARGET,
                       './' + Path(source).n)
        result.append((source, target))
    return result


@init.command()
def ext():
    """
    Copy external files.
    """
    inic = config.get(KEYS.INIT)
    tasks = inic.get(INIT_KEYS.EXTERNAL)
    import shutil
    import json
    results = []
    for t in external_to_copy():
        source = t[0]
        target = t[1]
        shutil.copyfile(source, target)
        results.append({INIT_KEYS.EXTERNAL_KEYS.SOURCE: source,
                        INIT_KEYS.EXTERNAL_KEYS.TARGET: target})
    results = json.dumps(results, indent=4, separators=(',', ': '))
    click.echo(results)


@init.command()
@click.option('--mac-auto', 'mac', flag_value='auto', default=True)
@click.option('--mac-no-create', 'mac', flag_value='no_create', default=True)
@click.option('--mac-force-create', 'mac', flag_value='force_create', default=True)
def auto():
    pass


# @click.command()
# @click.option('--config', '-c', type=str, default=c['pygate_config'], help='config YAML file name')
# @click.option('--pre', '-p', 'content', flag_value='pre', help='Initialize to pre make sub dirs.')
# @click.option('--dir', '-d', 'content', flag_value='sub', help='Make sub dirs')
# @click.option('--all', '-a', 'content', flag_value='all',  default=True, help='All tasks above')
# def init(config, content):
#     """
#     Initialize working directory
#     """
#     c = _load_config(config)
#     make_all = False
#     service.init(c, content)



class PhantomBinFileMaker:
    def __init__(self, fs, phantom_files):
        self.fs = fs
        self.phantom_files = phantom_files

    def _load_data(self, source):
        from dxpy.filesystem import Path
        import numpy as np
        from scipy.io import loadmat
        source = Path(source)
        source_syspath = self.fs.getsyspath(str(source))
        if source.suffix == '.npy':
            data = np.load(source_syspath)
            data = data.T
        else:
            data_dct = loadmat(source_syspath)
            data = data_dct[data_dct.keys()[0]]
        return data

    def _make_bin(self, target, data):
        import numpy as np
        with self.fs.open(target, 'wb') as fout:
            data = data.astype(np.uint16)
            fout.write(data.tostring())

    def make(self):
        for s, t in self.phantom_files:
            self._make_bin(t, self._load_data(s))



from pygate.components.source import *

# voxelized gamma(back to back) source


def voxelized_gamma(position, src_name='voxelized_gamma', read_table='act_range.dat', read_file='act.h33'):
    p = ParticleGamma()
    ang = AngularISO()
    v = Voxelized('act_range.dat', 'act.h33', position=position)
    src = Source(name=src_name, particle=p, angle=ang, shape=v)
    src_list = SourceList([src, ])
    return src_list

# voxelized


def voxelized_F18(position, src_name='voxelized_F18', read_table='act_range.dat', read_file='act.h33'):
    p = ParticlePositron()
    ang = AngularISO()
    v = Voxelized('act_range.dat', 'act.h33', position)
    src = Source(src_name, p, angle=ang, shape=v)
    src_list = SourceList([src, ])
    return src_list


def cylinder_source(position=Vec3(0, 0, 0), src_name='cylinder_source', cylinder=None,
                    activity=None, particle=None, angle=None):

    if cylinder is None:
        cylinder = Cylinder(5, 5, 'Volume')
    if activity is None:
        activity = 10000
    if particle is None:
        particle = ParticleGamma()
    if angle is None:
        angle = AngularISO()
    src = Source(src_name, particle, activity, angle, cylinder, position)
    src_list = SourceList([src, ])
    return src_list


def plane_source(position=Vec3(0, 0, 0), src_name='plane_source', rectangle=None, activity=None, particle=None, angle=None):
    if rectangle is None:
        rectangle = Rectangle([15, 15])
    if activity is None:
        activity = 1000
    if particle is None:
        particle = ParticleGamma(back2back=False)
    if angle is None:
        # default to the positive x direction.
        angle = AngularISO([90, 90, 0, 0])
    src = Source(src_name, particle, activity, angle, rectangle, position)
    src_list = SourceList([src, ])
    return src_list


def sphere_source(position=Vec3(0, 0, 0), src_name='sphere_source', sphere=None, activity=None, particle=None, angle=None):
    if sphere is None:
        sphere = Sphere(0.1, dimension='Volume')
    if activity is None:
        activity = 1000
    if particle is None:
        particle = ParticleGamma(back2back=False)
    if angle is None:
        # default to the positive x direction.
        angle = AngularISO([90, 90, 0, 0])
    src = Source(src_name, particle, activity, angle, sphere, position)
    src_list = SourceList([src, ])
    return src_list


def make_default_source(simu_name):
    if simu_name in ['OpticalSystem', 'OpticalGamma']:
        return sphere_source()
    else:
        return cylinder_source()


def sphere(radius,
           position: Vec3=Vec3(0.0, 0.0, 0.0, 'mm'),
           angle: Angular=None,
           activity: int=1000,
           particle=None,
           name='sphere_source') -> SourceList:
    """
    Parameters:
        - `radius`: radius of source
        - `position`:  
        - `angle`: default to the positive x direction
        - `activity`:

    Returns:
        - source list
    """
    sphere = Sphere(radius, dimension='Volume')
    if position is None:
        position = Vec3(0.0, 0.0, 0.0, 'mm')
    if particle is None:
        particle = ParticleGamma(unstable=False, halflife=None, back2back=False)
    if angle is None:
        angle = AngularISO([90, 90, 0, 0])
    return SourceList((Source(name, particle, activity, angle, sphere, position),))



from enum import Enum
from ..components.simulation import *
from ..components.geometry.volume import *
from ..components.geometry.camera import Camera
from ..components.geometry.phantom import Phantom
from ..components.geometry.surface import *
from ..components.physics import Cuts
from .cameras import *
from .phantoms import *
from .physics import *
from .digitizers import *
from .sources import *
from .parameters import *

# for reference
# simu_list = ['PETscanner','cylindricalPET','ecat','multiPatchPET','SPECThead','OpticalSystem','OpticalGamma']


class PredefinedSimulations(Enum):
    PETScanner = 'PETscanner'
    CylindricalPET = 'cylindricalPET'
    ECAT = 'ecat'
    MultiPatchPET = 'multiPatchPET'
    SPECTHead = 'SPECThead'
    OpticalSystem = 'OpticalSystem'
    OpticalGamma = 'OpticalGamma'


def make_default_camera(simu_name, world: Volume):
    if simu_name is 'cylindricalPET':
        return cylindricalPET(world)
    elif simu_name is 'ecat':
        return ecat(world)
    elif simu_name in ['OpticalSystem', 'OpticalGamma']:
        return opticalsystem(world)
    elif simu_name is 'multiPatchPET':
        return multipatchPET(world)
    else:
        raise ValueError("invalid simulation type: {}".format(simu_name))


def make_default_surfaces(simu_name, cam: Camera):
    surfaces = []
    if simu_name in ['OpticalSystem', 'OpticalGamma']:
        surfaces = optical_surfaces(cam)
    else:
        surfaces = ()
    return surfaces


def make_default_physics(simu_name, cam: Camera, phan: Phantom, cut_pair_list=None):
    # decide the cut pair list
    if cut_pair_list is None:
        cut_pair_list = []
        for v in cam.sds:
            cut_pair_list.append(Cuts(v, 0.1))
        if phan:
            for v in phan.sds:
                cut_pair_list.append(Cuts(v, 0.1))
        cut_pair_list = tuple(cut_pair_list)
    if simu_name in ['PETscanner', 'cylindricalPET', 'ecat', 'multiPatchPET']:
        return pet_physics(cut_pair_list)
    elif simu_name is 'SPECThead':
        return spect_physics(cut_pair_list)
    elif simu_name is 'OpticalSystem':
        return optical_physics(cut_pair_list=())
    elif simu_name is 'OpticalGamma':
        return gamma_physics(cut_pair_list=())
    else:
        raise ValueError(
            "simulation<set_physics> invalid system name: {}".format(simu_name))


def make_default_digitizer(simu_name, cam: Camera):
    if simu_name is 'cylindricalPET':
        return cylindricalPET_digitizer()
    elif simu_name is 'ecat':
        return ecat_digitizer(dtvolume=cam.system.levels['block'])
    elif simu_name in ['OpticalSystem', 'OpticalGamma']:
        return optical_digitizer()
    pass


def make_default_parameter(simu_name):
    if simu_name is 'OpticalSystem' or 'OpticalGamma':
        return optical_parameters()
    elif simu_name in ['PETscanner', 'ecat', 'cylindricalPET', 'multiPatchPET']:
        return pet_parameters()
    else:
        pass


def make_simulation(simu_name, geo=None, phy=None, digi=None, src=None, para=None):
       # for reference
    # simu_list = ['PETscanner','cylindricalPET','ecat','multiPatchPET','SPECThead','OpticalSystem','OpticalGamma']
    # simu_name = 'OpticalGamma'

   # must be given parts
    #####################
    #####################

    if geo is None:
        world = Box(name='world', size=Vec3(400, 400, 400, 'cm'))
        cam = make_default_camera(simu_name, world)
        surf = make_default_surfaces(simu_name, cam)
        phan = None
        # build up the geometry
        geo = Geometry(world, cam, phan, surf)

    # optional parts
    #####################
    #####################
    if phy is None:
        phy = make_default_physics(simu_name, cam, phan)
    if digi is None:
        digi = make_default_digitizer(simu_name, cam)

    # define the source
    if src is None:
        src = make_default_source(simu_name)

    # define the parameters
    if para is None:
        para = make_default_parameter(simu_name)

    # create a simulation
    simu = Simulation(geo, phy, digi, src, para)

    return simu


def simulation(simulation_name: PredefinedSimulations, geometry=None, physics=None, digitizer=None, source=None, parameter=None):
    return make_simulation(simulation_name.value, geometry, physics, digitizer, source, parameter)


def optical_gamma(source=None,
                  world_size: Vec3=Vec3(400.0, 400.0, 400.0, 'cm'),
                  crystal_size: Vec3=Vec3(30.0, 30.0, 30.0, 'mm'),
                  crystal_position: Vec3=Vec3(0.0, 0.0, 0.0, 'mm'),
                  crystal_material: str='LYSO',
                  nb_primaries=10000):
    from ..components.geometry import Box, Geometry
    from . import cameras, digitizers, physics, parameters
    world = Box('world', world_size)

    def crystal(mother): return Box('crystal', crystal_size,
                                    crystal_material, mother,
                                    crystal_position)
    return Simulation(Geometry(world, cameras.optical_gamma(world, crystal), None, ()),
                      physics.gamma_physics(()),
                      digitizers.optical_digitizer(),
                      source,
                      parameters.optical_gamma(nb_primaries))



from .api import *
"""
======
Pygate 
======

--------------
Pre simulation
--------------

Mac generator
=============

python API
----------

Low level API
^^^^^^^^^^^^^

High level API
^^^^^^^^^^^^^^

CLI
---

. code :: shell
    pygate generate mac --script experiment.py
    pygate generate mac --config experiment.yml
    pygate generate mac --config experiment.json


Shell script generator
======================


"""


class Submitter:
    def __init__(self, fs, config):
        from .utils import sub_dir_filters
        self.fs = fs
        self.c = config['submit']
        self.sub_dirs = sub_dir_filters(config)
        self.map_file = config['shell']['map']['output']
        self.merge_file = config['shell']['merge']['output']

    def _get_map_info(self):
        from dxpy.batch import DirectoriesFilter
        return [(self.fs.getsyspath(d), self.map_file) for d in DirectoriesFilter(self.sub_dirs).lst(self.fs)]

    def _get_merge_info(self):
        return (self.fs.getsyspath('.'), self.merge_file)

    def submit(self):
        run_infos = self._get_map_info()
        merge_infos = self._get_merge_info()
        if self.c['worker'] == 'slurm':
            self._slurm(run_infos, merge_infos)
        elif self.c['worker'] == 'hqlf':
            self._hqlf(run_infos, merge_infos)

    def _echo(self, info, fout):
        msg = '\t'.join([str(e) for e in info])
        print(msg)
        print(msg, file=fout)

    def _slurm(self, run_infos, post_infos):
        from dxpy.slurm import sbatch
        sids = []
        with self.fs.open(self.c['output'], 'w') as fout:
            for t in run_infos:
                sid = sbatch(t[0], t[1])
                self._echo(('MAP', sid, t[0], t[1]), fout)
                sids.append(sid)
            sids = [str(s) for s in sids]
            deparg = '--dependency=afterok:' + ':'.join(sids)
            sid = sbatch(post_infos[0], post_infos[1], deparg)
            self._echo(('MERGE', sid, post_infos[0], post_infos[1]), fout)

    def _hqlf(self, run_infos, post_infos):
        from dxpy.task.model import creators
        from dxpy.task import interface
        tasks = []
        for i, t in enumerate(run_infos):
            desc = '<PYGATE {0}>.RUN #{1}'.format(post_infos[0], i)
            tasks.append(creators.task_slurm(
                file=t[1], workdir=t[0], desc=desc))
        tasks.append(creators.task_slurm(file=post_infos[1],
                                         workdir=post_infos[0],
                                         desc='<PYGATE {0}>.POST'.format(post_infos[0])))
        depens = [None] * (len(tasks) - 1)
        depens.append(list(range(len(tasks) - 1)))
        g = creators.task_graph(tasks, depens)
        # print('\n'.join([t.to_json() for t in tasks]))
        ids = interface.create_graph(g)
        print('Submitted to HQLF.Task with tids:', ids)
        with self.fs.open(self.c['output'], 'w') as fout:
            print('Submitted via HQLF, with tids:', file=fout)
            print(ids, file=fout)



from .base import ObjectWithTemplate
class Vec3(ObjectWithTemplate):
    template = 'vec3'

    def __init__(self, x, y, z, unit=None):
        self.x = x
        self.y = y
        self.z = z
        self.unit = unit



class ShellScriptBase:
    def __init__(self, fs, workdir: str, output: str, tasks: list, shell='zsh', version='7.2'):
        self.fs = fs
        self.workdir = workdir
        self.output = output
        self.tasks = tasks
        self.shell = shell.lower()
        if not self.shell in ['zsh', 'bash']:
            raise ValueError("Unknown shell {}.".format(self.shell))
        self.content = None
        self.task_type = None
        self.version = version

    def _load_shell_template(self):
        from .utils import load_script
        import jinja2
        script_name = "{0}_{1}.sh".format(self.task_type, self.shell)
        return jinja2.Template(load_script(script_name))

    def _make(self):
        raise NotImplemented

    def dump(self):
        self._make()
        with self.fs.opendir(self.workdir) as d:
            with d.open(self.output, 'w') as fout:
                print(self.content, file=fout)


class ShellScriptMap(ShellScriptBase):
    def __init__(self, fs, workdir: str, output: str, tasks: list, shell='zsh', version="7.2"):
        super(__class__, self).__init__(fs, workdir, output, tasks, shell, version)
        self.task_type = 'map'
        self.commands = []

    def _add_if_gate(self, task):
        if task.get('method') == 'Gate':
            self.commands.append("Gate {}".format(task['filename']))

    def _add_if_root(self, task):
        if task['method'] == 'root':
            self.commands.append("root -q -b {}".format(task['filename']))

    def _get_workdir_on_local_and_server(self):
        from dxpy.filesystem import Path
        import os
        p_ser = self.fs.getsyspath(str(self.workdir))
        import random
        temps = random.randint(0, 100000000)
        subd = 'tmp{}'.format(temps)
        p_loc = Path('/tmp') / subd
        return p_ser, p_loc

    def _make(self):
        scrpt_tpl = self._load_shell_template()
        supported_tasks = [self._add_if_gate, self._add_if_root]
        for t in self.tasks:
            for add_if_func in supported_tasks:
                add_if_func(t)
        p_ser, p_loc = self._get_workdir_on_local_and_server()
        self.content = scrpt_tpl.render(local_work_directory=p_loc,
                                        server_work_directory=p_ser,
                                        commands='\n'.join(self.commands),
                                        version=self.version)


class ShellScriptMerge(ShellScriptBase):
    def __init__(self, fs, workdir: str, output: str, tasks: list, shell='zsh'):
        super(__class__, self).__init__(fs, workdir, output, tasks, shell)
        self.task_type = 'merge'

    def _make(self):
        scrpt_tpl = self._load_shell_template()
        self.content = scrpt_tpl.render(merge='merge' in self.tasks,
                                        clean='clean' in self.tasks)


class ShellScriptMaker:
    def __init__(self, fs, config):
        from .utils import sub_dir_filters
        self.fs = fs
        self.c = config['shell']
        self.c['version'] = config['version']
        self.sub_filters = sub_dir_filters(config)

    def make(self):
        from dxpy import batch
        sub_dirs = batch.DirectoriesFilter(self.sub_filters).lst(self.fs)
        for d in sub_dirs:
            ShellScriptMap(self.fs, d, self.c['map']['output'],
                           self.c['map']['tasks'], self.c['type'], self.c['version']).dump()
        ShellScriptMerge(self.fs, '.', self.c['merge']['output'],
                         self.c['merge']['tasks'], self.c['type']).dump()



import click


def auto_sub():
    from ..conf import config, KEYS
    if config.get(KEYS.SUB_PATTERNS) is None:
        config[KEYS.SUB_PATTERNS] = [config.get(KEYS.SUB_PREFIX) + '*']
        config[KEYS.SUB_FORMAT] = config.get(KEYS.SUB_PREFIX) + r'.{}'


def load_config(filename, is_no_config, dryrun):
    from ..conf import config, KEYS
    import yaml
    import json
    if dryrun is not None:
        config['dryrun'] = dryrun
    if not is_no_config and filename is not None:
        with open(filename, 'r') as fin:
            if filename.endswith('yml'):
                config.update(yaml.load(fin))
            else:
                config.update(json.load(fin))
    auto_sub()


@click.group()
@click.option('--config', '-c', help="config file name", default=None)
@click.option('--no-config', help="ignore config file", is_flag=True)
@click.option('--dryrun', help='Do not do anything, just show expected results.', is_flag=True)
def pygate(config, no_config, dryrun):
    from dxl.fs import File
    if config is None:
        f = File('./pygate.yml')
        if f.exists():
            config = f.path.n
    if config is None:
        f = File('./pygate.json')
        if f.exists():
            config = f.path.n
    load_config(config, no_config, dryrun)


from .clean import clean
from .initialize import init
from .analysis import analysis
from .merge import merge
from .submit import submit
# class CLI(click.MultiCommand):
#     commands = {'init': None,
#                 'merge': None,
#                 'clean': None,
#                 'analysis': None,
#                 'submit': None}

#     def __init__(self):
#         super(__class__, self).__init__(name='pygate', help=__class__.__doc__)

#     def list_commands(self, ctx):
#         return sorted(self.commands.keys())

#     def get_command(self, ctx, name):
#         from .initialize import init
#         from .analysis import analysis
#         from .merge import merge
#         from .clean import clean
#         from .submit import submit
#         cmd_map = {
#             'init': init,
#             'merge': merge,
#             'clean': clean,
#             'submit': submit,
#             'analysis': analysis
#         }
#         if name in self.commands and self.commands[name] is None:
#             self.commands[name] = cmd_map[name]
#         return self.commands.get(name)

cli = pygate(obj={})

# pygate = CLI()



from setuptools import setup, find_packages

setup(name='dxl-pygate',
      version='0.12.6',
      description='A simplified python interface for Gate.',
      url='https://github.com/Hong-Xiang/pygate',
      author='Hong Xiang',
      author_email='hx.hongxiang@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=['fs', 'click',
                        'rx', 'jinja2', 'dxl-dxpy>=0.5',
                        'dxl-fs', 'dxl-cluster', 'dask'],
      entry_points='''
            [console_scripts]
            pygate=pygate.cli:cli
      ''',
      include_package_data=True,
      zip_safe=False)



from fs.osfs import OSFS
from .template.service import MacMaker
from .utils import load_script
from .config import config as c


class ConfigMaker:
    @classmethod
    def _make_pygate_config(cls, fs, target, config_filename=None):
        config_filename = config_filename or c['pygate_config']
        with fs.opendir(target) as d:
            with d.open(config_filename, 'w') as fout:
                print(load_script(config_filename), file=fout)

    @classmethod
    def _make_mac_config(cls, fs, target, config_filename: str=None):
        if config_filename.endswith('.yml'):
            config_filename = config_filename[:-4]
        MacMaker.make_yml(config_filename)

    @classmethod
    def make(cls, fs, target, pygate_config=None, mac_config=None):
        pygate_config = pygate_config or c['pygate_config']
        mac_config = mac_config or c['mac_config']
        cls._make_pygate_config(fs, target, pygate_config)
        cls._make_mac_config(fs, target, mac_config)



from enum import Enum
from functools import reduce
from typing import Tuple

import numpy as np
import pandas as pd
import dask.dataframe as ddf
import dask
from dask.dataframe.groupby import DataFrameGroupBy


class ParticleID(Enum):
    Gamma = 22


class ColumnNames:
    Event = 'eventID'
    Process = 'processName'
    Particle = 'particalID'
    SourceX = 'srcX'
    SourceY = 'srcY'
    SourceZ = 'srcZ'
    X = 'x'
    Y = 'y'
    Z = 'z'
    EnergyDeposit = 'energy'


class ResultBase:
    def __init__(self, data=None):
        self.data = data

    @property
    def d(self):
        return self.data


class Results(ResultBase):
    def __init__(self, data: Tuple[ResultBase]):
        super().__init__(tuple(data))

    def map(self, func) -> 'Results':
        return Results(map(func, self.d))

    def filter(self, func) -> 'Results':
        return Results(filter(func, self.d))

    def call(self, func_name) -> 'Results':
        return Results(map(self.d, lambda x: getattr(x, 'func_name')()))

    def zip(self, r: 'Results'):
        return Results(zip(self.d, r.d))

    def first(self) -> ResultBase:
        return self.d[0]

    def flatten(self) -> 'Results':
        if not isinstance(self.d[0], Results):
            raise TypeError(
                "Flatten requires Results of Results, got {}.".format(type(self.d[0])))
        return Results(tuple(sum([list(d.d) for d in self.d], [])))

    def merge(self) -> ResultBase:
        result = None
        for o in self.d:
            if result is None:
                result = o
            else:
                result = result.merge(o)
        return result

    def to_list(self):
        return list(self.d)

class ResultsDask(ResultBase):
    def map(self, func) -> 'ResultsDask':
        return Results(dask.delayed(map)(func, self.d))
    


class ResultsNamedTuple(ResultBase):
    pass




class ResultsWithKeys(Results):
    def __init__(self, data: Tuple[Tuple[str, ResultBase]]):
        super().__init__(data)

    def drop_keys(self) -> Results:
        return self.map(lambda x: x[1])

    def to_dict(self):
        return {o[0]: o[1] for o in self.d}

    def select(self, key):
        for o in self.d:
            if o[0] == key:
                return o[1]
        raise KeyError("Key {} not found.".format(key))


class ResultsWithUnknownKeys(ResultsDask):
    def select(self, key):
        return DataFrame(self.d.get_group(key))

    def drop_keys(self) -> ResultsDask:
        return self.map(lambda x: x[1])


class Series(ResultBase):
    def __init__(self, series: pd.Series):
        super().__init__(series)

    def columns(self, *cols):
        return (self.d[c] for c in cols)

    def position(self) -> 'Vec3':
        return Vec3(*self.columns(ColumnNames.X,
                                  ColumnNames.Y,
                                  ColumnNames.Z))

    def source_position(self) -> 'Vec3':
        return Vec3(*self.columns(ColumnNames.SourceX,
                                  ColumnNames.SourceY,
                                  ColumnNames.SourceZ))

    def energy_deposit(self) -> float:
        return self.d[ColumnNames.EnergyDeposit]


class DataFrame(ResultBase):
    def __init__(self, dataframe: pd.DataFrame):
        super().__init__(dataframe)

    def first(self, *columns) -> Series:
        return Series(self.d.iloc[0])

    def split_by(self, column: str) -> ResultsWithKeys:
        groups = self.d.groupby(column)
        if isinstance(groups, DataFrameGroupBy):
            return ResultsWithUnknownKeys(groups)
        else:
            return ResultsWithKeys(((k, DataFrame(groups.get_group(k).drop([column], axis=1))) for k in groups.groups))

    def split_row(self) -> Results:
        return Results((Series(self.d.iloc[i]) for i in range(self.d.shape[0])))

    def merge(self, r) -> 'DataFrame':
        if not isinstance(r, DataFrame):
            raise TypeError(
                "Can not merge {} with {}.".format(__class__, type(r)))
        return DataFrame(pd.concat([self.d, r.d], axis=0))

    def to_event(self) -> 'Event':
        if not ColumnNames.Event in self.d.columns:
            return Event(self.d)




class Vec3(ResultBase):
    def __init__(self, x, y=None, z=None):
        super().__init__(np.array([x, y, z]))

    @property
    def x(self) -> float:
        return self.d[0]

    @property
    def y(self) -> float:
        return self.d[1]

    @property
    def z(self) -> float:
        return self.d[2]

    def to_list(self):
        return tuple(self.d)


class EnergyDeposit(ResultBase):
    def __init__(self, position, energy):
        super().__init__((position, energy))

    @property
    def position(self):
        return self.d[0]

    @property
    def energy(self):
        return self.d[1]


class Event(DataFrame):
    def __init__(self, dataframe: pd.DataFrame):
        super().__init__(dataframe)

    def source_position(self) -> Vec3:
        return self.first().source_position()

    def first_position(self) -> Vec3:
        return self.first().position()

    def incident_direction(self) -> Vec3:
        dp = self.first_position().d - self.source_position().d
        return Vec3(*tuple(dp / np.linalg.norm(dp)))

    def energy_deposit_list(self) -> Results:
        return self.split_row().map(lambda s: EnergyDeposit(s.position(), s.energy_deposit()))


class CSVFile(ResultBase):
    def __init__(self, filename: str):
        super().__init__(filename)

    def load(self) -> DataFrame:
        return DataFrame(pd.read_csv(self.d))



from ..base import ObjectWithTemplate
from .volume import Volume
from typing import Tuple
from .surface import Surface


class GeometryPreInitialise(ObjectWithTemplate):
    template = 'geometry/geometry_pre_init'

    def __init__(self, world: Volume,
                 camera, phantom):
        self.world = world
        self.camera = camera
        self.phantom = phantom


class GeometryPostInitialise(ObjectWithTemplate):
    template = 'geometry/geometry_post_init'

    def __init__(self, surfaces):
        self.surfaces = surfaces


class Geometry:
    def __init__(self, world: Volume,
                 camera, phantom,
                 surfaces: Tuple[Surface]=()):
        self.geometry_pre_initialise = GeometryPreInitialise(world,
                                                             camera, phantom)
        self.geometry_post_initialise = GeometryPostInitialise(surfaces)

    def render_pre(self):
        return self.geometry_pre_initialise.render()

    def render_post(self):
        return self.geometry_post_initialise.render()



# _*_ coding: utf-8 _*_
from math import pi
import yaml

from ..base import ObjectWithTemplate
from ..utils import Vec3
from typing import Tuple




class Repeater(ObjectWithTemplate):
    template = 'geometry/volume/repeater/repeater'
    repeater_type = None


class RepeaterRing(Repeater):
    template = 'geometry/volume/repeater/ring'
    repeater_type = 'ring'

    def __init__(self, number):
        super().__init__()
        self.n = number


class RepeaterLinear(Repeater):
    template = 'geometry/volume/repeater/linear'
    repeater_type = 'linear'

    def __init__(self, number, repeat_vector):
        super().__init__()
        self.n = number
        self.rv = repeat_vector


class RepeaterCubic(Repeater):
    template = 'geometry/volume/repeater/cubic'
    repeater_type = 'cubicArray'

    def __init__(self, scale: Vec3, repeat_vector: Vec3):
        super().__init__()
        self.scale = scale
        self.rv = repeat_vector


class Volume(ObjectWithTemplate):
    shape_type = 'volume'
    template = 'geometry/volume/volume'

    def __init__(self, name, material=None, mother=None, position=None, unit=None, repeaters: Repeater=None):
        self.mother = mother
        if self.mother is not None:
            self.mother.add_child(self)
        self.name = name
        self.material = material
        self.position = position
        self.unit = unit or 'mm'
        if self.position is not None and self.position.unit is None:
            self.position.unit = self.unit
        self.repeaters = repeaters or ()
        for r in self.repeaters:
            r.volume = self
        self.children = []

    def add_child(self, child):
        if not child.mother is self:
            raise ValueError(
                "Trying to ({}).add_child({}) with another mother {}.".format(self.name, child.name, child.mother.name))
        child.mother = self
        if not child in self.children:
            self.children.append(child)
        return child


class Box(Volume):
    shape_type = 'box'
    template = 'geometry/volume/box'

    def __init__(self, name, size, material=None, mother=None, position=None, unit=None, repeaters: Repeater=None):
        super().__init__(name, material, mother, position, unit, repeaters)
        self.size = size
        if self.size.unit is None:
            self.size.unit = self.unit


class Cylinder(Volume):
    shape_type = 'cylinder'
    template = 'geometry/volume/cylinder'

    def __init__(self, name, rmax, rmin=None, height=None,
                 phi_start=None, delta_phi=None,
                 material=None, mother=None, position=None, unit=None, repeaters: Repeater=None):
        super().__init__(name, material, mother, position, unit, repeaters)
        self.rmax = rmax
        self.rmin = rmin
        self.height = height
        self.phi_start = phi_start
        self.delta_phi = delta_phi


class Sphere(Volume):
    template = 'geometry/volume/sphere'
    shape_type = 'sphere'

    def __init__(self, name, rmax, rmin=None,
                 phi_start=None, delta_phi=None,
                 theta_start=None, delta_theta=None,
                 material=None, mother=None, position=None, unit=None, repeaters: Repeater=None):
        super().__init__(name, material, mother, position, unit, repeaters)
        self.rmax = rmax
        self.rmin = rmin
        self.phi_start = phi_start
        self.delta_phi = delta_phi
        self.theta_start = theta_start
        self.delta_theta = delta_theta


class ImageRegularParamerisedVolume(Volume):
    template = 'geometry/volume/image_volume'
    shape_type = 'ImageRegularParametrisedVolume'

    def __init__(self,  name, image_file, range_file,
                 material=None, mother=None, position=None, unit=None, repeaters: Repeater=None):
        super().__init__(name, material, mother, position, unit, repeaters)
        self.image_file = image_file
        self.range_file = range_file


class Patch(Volume):
    template = 'geometry/patch'
    shape_type = 'shape'

    def __init__(self, name, patch_file, material=None, mother=None, position=None, unit=None, repeater: Repeater = None):
        super().__init__(name, material, mother, position, unit, repeater)
        self.patch_file = patch_file








import click
from ...config import config as c
from ... import service


def _load_config(config):
    import yaml
    with open(config) as fin:
        return yaml.load(fin)


@click.command()
def make_config():
    service.make_config()


@click.command()
@click.option('--config', '-c', type=str, default=c['pygate_config'], help='config YAML file name')
@click.option('--pre', '-p', 'content', flag_value='pre', help='Initialize to pre make sub dirs.')
@click.option('--dir', '-d', 'content', flag_value='sub', help='Make sub dirs')
@click.option('--all', '-a', 'content', flag_value='all',  default=True, help='All tasks above')
def init(config, content):
    """
    Initialize working directory
    """
    c = _load_config(config)
    make_all = False
    service.init(c, content)


@click.command()
@click.option('--config', '-c', type=str, default=c['pygate_config'], help='config YAML file name')
def submit(config):
    # TODO: add submit service
    c = _load_config(config)
    service.submit(c)


@click.command()
@click.option('--config', '-c', type=str, default=c['pygate_config'], help='config YAML file name')
def merge(config):
    c = _load_config(config)
    service.merge(c)


@click.command()
@click.option('--config', '-c', type=str, default=c['pygate_config'], help='config YAML file name')
@click.option('--dirs', '-d', 'content', flag_value='dirs')
@click.option('--sources', '-s', 'content', flag_value='sources')
@click.option('--all', '-a', 'content', flag_value='all')
@click.option('--dryrun', is_flag=True)
def clean(config, content, dryrun):
    c = _load_config(config)
    if content == 'dirs':
        c['clean']['sub_dirs'] = True
        c['clean']['sources'] = False
    elif content == 'sources':
        c['clean']['sources'] = True
        c['clean']['sub_dirs'] = False
    elif content == 'all':
        c['clean']['sources'] = True
        c['clean']['sub_dirs'] = True
    service.clean(c)



from typing import List
from .base import ObjectWithTemplate


class Acquisition(ObjectWithTemplate):
    template = 'parameter/acquisition/acquisition'
    acquisition_type = None


class AcquisitionPrimaries(Acquisition):
    template = 'parameter/acquisition/primaries'
    acquisition_type = 'primaries'

    def __init__(self, number=10000):
        super().__init__()
        self.number = number


class AcquisitionPeriod(Acquisition):
    template = 'parameter/acquisition/period'
    acquisition_type = 'period'

    def __init__(self, start=0.0, end=1.0, step=1.0):
        super().__init__()
        self.start = start
        self.step = step
        self.end = end


class Output(ObjectWithTemplate):
    template = 'parameter/output/output'
    output_type = None

    def __init__(self, file_name):
        self.file_name = file_name


class Ascii(Output):
    template = 'parameter/output/ascii'
    output_type = 'ascii'

    def __init__(self, file_name, hit=0, singles=0, coincidences=0):
        super().__init__(file_name)
        self.hit = hit
        self.coincidences = coincidences
        self.singles = singles


class Binary(Output):
    template = 'parameter/output/binary'
    output_type = 'binary'

    def __init__(self, file_name, hit=0, singles=0, coincidences=0):
        super().__init__(file_name)
        self.hit = hit
        self.coincidences = coincidences
        self.singles = singles


class Root(Output):
    template = 'parameter/output/root'
    output_type = 'root'

    def __init__(self, file_name,
                 hit=None, singles=None, coincidences=None, optical=None, delay=None):
        super().__init__(file_name)
        self.hit = hit
        self.singles = singles
        self.coincidences = coincidences
        self.optical = optical
        self.delay = delay


class Sinogram(Output):
    template = 'parameter/output/sinogram'
    output_type = 'sinogram'

    def __init__(self, file_name,
                 input_,
                 radial_bin=None,
                 is_true_only=None,
                 is_raw_output=None,
                 tang_blurring=None,
                 axial_blurring=None,
                 is_store_delay=None,
                 is_store_scatter=None):
        super().__init__(file_name)
        self.input_ = input_
        self.radial_bin = radial_bin
        self.is_true_only = is_true_only
        self.is_raw_output = is_raw_output
        self.tang_blurring = tang_blurring
        self.axial_blurring = axial_blurring
        self.is_store_delay = is_store_delay
        self.is_store_scatter = is_store_scatter


class RandomEngine(ObjectWithTemplate):
    template = 'parameter/random_engine/random_engine'
    engine_type = 'JamesRandom'

    def __init__(self, seed='default'):
        self.seed = seed


class RandomEngineRanlux64(RandomEngine):
    engine_type = 'Ranlux64'


class RandomEngineJamesRandom(RandomEngine):
    engine_type = 'JamesRandom'


class RandomEngineMersenneTwister(RandomEngine):
    engine_type = 'MersenneTwister'


class Parameter(ObjectWithTemplate):
    template = 'parameter/parameter'

    def __init__(self, random_engine, outputs: List[Output], acquisition):
        self.acquisition = acquisition
        self.outputs = outputs
        self.random_engine = random_engine



from pygate.components.geometry.volume import *
from pygate.components.geometry.phantom import Phantom


def voxelized_phantom(world, image_file = 'phan.h33', range_file = 'mat_range.dat', position = None):
    
    iv = ImageRegularParamerisedVolume(name='voxelized_phantom', image_file = image_file,
                                                  range_file = range_file, mother=world, position=position)
    phan = Phantom([iv,])
    return phan



class Cleaner:
    def __init__(self, fs, config):
        from .utils import sub_dir_filters
        self.fs = fs
        self.c = config['clean']
        self.sub_dir = sub_dir_filters(config)

    def clean(self):
        if self.c['sub_dirs']:
            self._clean_subs()
        if self.c['sources']:
            self._clean_sources()

    def msg(self, path):
        from fs.path import normpath
        if self.c['verbose'] > 0:
            print('DELETE: {}'.format(normpath(self.fs.getsyspath(path))))

    def _clean_subs(self):
        from dxpy.batch import DirectoriesFilter
        dirs = DirectoriesFilter(self.sub_dir).lst(self.fs)
        for d in dirs:
            self.msg(d)
            if not self.c['dryrun']:
                self.fs.removetree(d)

    def _clean_sources(self):
        from dxpy.batch import FilesFilter
        files = FilesFilter(['*']).lst(self.fs)
        for f in files:
            if f in self.c['keep']:
                continue
            self.msg(f)
            if not self.c['dryrun']:
                self.fs.remove(f)
